import pytest

from message_ix_models.report.config import Config


class TestConfig:
    def test_use_file(self, tmp_path):
        cfg = Config()

        # No effect
        cfg.use_file(None)

        # Passing a missing path raises an exception
        with pytest.raises(
            FileNotFoundError, match="Reporting configuration in .*missing"
        ):
            cfg.use_file(tmp_path.joinpath("missing"))

        # Passing a file name that does not exist raises an exception
        with pytest.raises(
            FileNotFoundError, match=r"Reporting configuration in 'unknown\(.yaml\)'"
        ):
            cfg.use_file("unknown")



"""Prepare data for water use for cooling & energy technologies."""

import numpy as np
import pandas as pd
from message_ix import make_df

from message_ix_models.model.water.data.demands import read_water_availability
from message_ix_models.model.water.utils import map_yv_ya_lt
from message_ix_models.util import broadcast, package_data_path, same_node, same_time


def map_basin_region_wat(context):
    """
    Calculate share of water avaialbility of basins per each parent region.

    The parent region could be global message regions or country

    Parameters
    ----------
    context : .Context
    Returns
    -------
    data : dict of (str -> pandas.DataFrame)
        Keys are MESSAGE parameter names such as 'input', 'fix_cost'. Values
        are data frames ready for :meth:`~.Scenario.add_par`.
    """
    info = context["water build info"]

    if "year" in context.time:
        PATH = package_data_path(
            "water", "delineation", f"basins_by_region_simpl_{context.regions}.csv"
        )
        df_x = pd.read_csv(PATH)
        # Adding freshwater supply constraints
        # Reading data, the data is spatially and temprally aggregated from GHMs
        path1 = package_data_path(
            "water",
            "availability",
            f"qtot_5y_{context.RCP}_{context.REL}_{context.regions}.csv",
        )

        df_sw = pd.read_csv(path1)
        df_sw.drop(["Unnamed: 0"], axis=1, inplace=True)

        # Reading data, the data is spatially and temporally aggregated from GHMs
        df_sw["BCU_name"] = df_x["BCU_name"]

        if context.type_reg == "country":
            df_sw["MSGREG"] = context.map_ISO_c[context.regions]
        else:
            df_sw["MSGREG"] = f"{context.regions}_" + df_sw["BCU_name"].str[-3:]

        df_sw = df_sw.set_index(["MSGREG", "BCU_name"])

        # Calculating ratio of water availability in basin by region
        df_sw = df_sw.groupby(["MSGREG"]).apply(lambda x: x / x.sum())
        df_sw.reset_index(inplace=True)
        df_sw["Region"] = "B" + df_sw["BCU_name"].astype(str)
        df_sw["Mode"] = df_sw["Region"].replace(regex=["^B"], value="M")
        df_sw.drop(columns=["BCU_name"], inplace=True)
        df_sw.set_index(["MSGREG", "Region", "Mode"], inplace=True)
        df_sw = df_sw.stack().reset_index(level=0).reset_index()
        df_sw.columns = ["region", "mode", "date", "MSGREG", "share"]
        df_sw.sort_values(["region", "date", "MSGREG", "share"], inplace=True)
        df_sw["year"] = pd.DatetimeIndex(df_sw["date"]).year
        df_sw["time"] = "year"
        df_sw = df_sw[df_sw["year"].isin(info.Y)]
        df_sw.reset_index(drop=True, inplace=True)

    else:
        # add water return flows for cooling tecs
        # Use share of basin availability to distribute the return flow from
        path3 = package_data_path(
            "water",
            "availability",
            f"qtot_5y_m_{context.RCP}_{context.REL}_{context.regions}.csv",
        )
        df_sw = pd.read_csv(path3)

        # reading sample for assiging basins
        PATH = package_data_path(
            "water", "delineation", f"basins_by_region_simpl_{context.regions}.csv"
        )
        df_x = pd.read_csv(PATH)

        # Reading data, the data is spatially and temporally aggregated from GHMs
        df_sw["BCU_name"] = df_x["BCU_name"]

        if context.type_reg == "country":
            df_sw["MSGREG"] = context.map_ISO_c[context.regions]
        else:
            df_sw["MSGREG"] = f"{context.regions}_" + df_sw["BCU_name"].str[-3:]

        df_sw = df_sw.set_index(["MSGREG", "BCU_name"])
        df_sw.drop(columns="Unnamed: 0", inplace=True)

        # Calculating ratio of water availability in basin by region
        df_sw = df_sw.groupby(["MSGREG"]).apply(lambda x: x / x.sum())
        df_sw.reset_index(inplace=True)
        df_sw["Region"] = "B" + df_sw["BCU_name"].astype(str)
        df_sw["Mode"] = df_sw["Region"].replace(regex=["^B"], value="M")
        df_sw.drop(columns=["BCU_name"], inplace=True)
        df_sw.set_index(["MSGREG", "Region", "Mode"], inplace=True)
        df_sw = df_sw.stack().reset_index(level=0).reset_index()
        df_sw.columns = ["node", "mode", "date", "MSGREG", "share"]
        df_sw.sort_values(["node", "date", "MSGREG", "share"], inplace=True)
        df_sw["year"] = pd.DatetimeIndex(df_sw["date"]).year
        df_sw["time"] = pd.DatetimeIndex(df_sw["date"]).month
        df_sw = df_sw[df_sw["year"].isin(info.Y)]
        df_sw.reset_index(drop=True, inplace=True)

    return df_sw


def add_water_supply(context):
    """Add Water supply infrastructure
    This function links the water supply based on different settings and options.
    It defines the supply linkages for freshwater, groundwater and salinewater.
    Parameters
    ----------
    context : .Context
    Returns
    -------
    data : dict of (str -> pandas.DataFrame)
        Keys are MESSAGE parameter names such as 'input', 'fix_cost'.
        Values are data frames ready for :meth:`~.Scenario.add_par`.
        Years in the data include the model horizon indicated by
        ``context["water build info"]``, plus the additional year 2010.
    """
    # define an empty dictionary
    results = {}

    # Reference to the water configuration
    info = context["water build info"]
    # load the scenario from context
    scen = context.get_scenario()

    year_wat = [2010, 2015]
    fut_year = info.Y
    year_wat.extend(info.Y)
    sub_time = context.time

    # first activity year for all water technologies is 2020
    first_year = scen.firstmodelyear

    print(" future year = ", fut_year)
    print(" year_wat = ", year_wat)

    # reading basin_delineation
    FILE = f"basins_by_region_simpl_{context.regions}.csv"
    PATH = package_data_path("water", "delineation", FILE)

    df_node = pd.read_csv(PATH)
    # Assigning proper nomenclature
    df_node["node"] = "B" + df_node["BCU_name"].astype(str)
    df_node["mode"] = "M" + df_node["BCU_name"].astype(str)
    if context.type_reg == "country":
        df_node["region"] = context.map_ISO_c[context.regions]
    else:
        df_node["region"] = f"{context.regions}_" + df_node["REGION"].astype(str)

    # Storing the energy MESSAGE region names
    node_region = df_node["region"].unique()

    # reading groundwater energy intensity data
    FILE1 = f"gw_energy_intensity_depth_{context.regions}.csv"
    PATH1 = package_data_path("water", "availability", FILE1)
    df_gwt = pd.read_csv(PATH1)
    if context.type_reg == "country":
        df_gwt["region"] = context.map_ISO_c[context.regions]
    else:
        df_gwt["REGION"] = f"{context.regions}_" + df_gwt["REGION"].astype(str)

    # reading groundwater energy intensity data
    FILE2 = f"historical_new_cap_gw_sw_km3_year_{context.regions}.csv"
    PATH2 = package_data_path("water", "availability", FILE2)
    df_hist = pd.read_csv(PATH2)
    df_hist["BCU_name"] = "B" + df_hist["BCU_name"].astype(str)

    if context.nexus_set == "cooling":
        # Add output df  for surfacewater supply for regions
        output_df = (
            make_df(
                "output",
                technology="extract_surfacewater",
                value=1,
                unit="km3",
                year_vtg=year_wat,
                year_act=year_wat,
                level="water_supply",
                commodity="freshwater",
                mode="M1",
                time="year",
                time_dest="year",
                time_origin="year",
            )
            .pipe(broadcast, node_loc=node_region)
            .pipe(same_node)
        )

        # Add output df  for groundwater supply for regions
        output_df = output_df.append(
            make_df(
                "output",
                technology="extract_groundwater",
                value=1,
                unit="km3",
                year_vtg=year_wat,
                year_act=year_wat,
                level="water_supply",
                commodity="freshwater",
                mode="M1",
                time="year",
                time_dest="year",
                time_origin="year",
            )
            .pipe(broadcast, node_loc=node_region)
            .pipe(same_node)
        )
        # Add output of saline water supply for regions
        output_df = output_df.append(
            make_df(
                "output",
                technology="extract_salinewater",
                value=1,
                unit="km3",
                year_vtg=year_wat,
                year_act=year_wat,
                level="water_supply",
                commodity="saline_ppl",
                mode="M1",
                time="year",
                time_dest="year",
                time_origin="year",
            )
            .pipe(broadcast, node_loc=node_region)
            .pipe(same_node)
        )
        results["output"] = output_df

    elif context.nexus_set == "nexus":
        # input data frame  for slack technology balancing equality with demands
        inp = (
            make_df(
                "input",
                technology="return_flow",
                value=1,
                unit="-",
                level="water_avail_basin",
                commodity="surfacewater_basin",
                mode="M1",
                year_vtg=year_wat,
                year_act=year_wat,
            )
            .pipe(
                broadcast,
                node_loc=df_node["node"],
                time=sub_time,
            )
            .pipe(same_node)
            .pipe(same_time)
        )

        # input data frame  for slack technology balancing equality with demands
        inp = inp.append(
            make_df(
                "input",
                technology="gw_recharge",
                value=1,
                unit="-",
                level="water_avail_basin",
                commodity="groundwater_basin",
                mode="M1",
                year_vtg=year_wat,
                year_act=year_wat,
            )
            .pipe(
                broadcast,
                node_loc=df_node["node"],
                time=sub_time,
            )
            .pipe(same_node)
            .pipe(same_time)
        )

        # input dataframe  linking water supply to energy dummy technology
        inp = inp.append(
            make_df(
                "input",
                technology="basin_to_reg",
                value=1,
                unit="-",
                level="water_supply_basin",
                commodity="freshwater_basin",
                mode=df_node["mode"],
                node_origin=df_node["node"],
                node_loc=df_node["region"],
            )
            .pipe(
                broadcast,
                year_vtg=year_wat,
                time=sub_time,
            )
            .pipe(same_time)
        )
        inp["year_act"] = inp["year_vtg"]
        # # input data frame  for slack technology balancing equality with demands
        # inp = inp.append(
        #     make_df(
        #         "input",
        #         technology="salinewater_return",
        #         value=1,
        #         unit="-",
        #         level="water_avail_basin",
        #         commodity="salinewater_basin",
        #         mode="M1",
        #         time="year",
        #         time_origin="year",
        #         node_origin=df_node["node"],
        #         node_loc=df_node["node"],
        #     ).pipe(broadcast, year_vtg=year_wat, year_act=year_wat)
        # )

        # input data frame  for freshwater supply
        yv_ya_sw = map_yv_ya_lt(year_wat, 50, first_year)

        inp = inp.append(
            make_df(
                "input",
                technology="extract_surfacewater",
                value=1,
                unit="-",
                level="water_avail_basin",
                commodity="surfacewater_basin",
                mode="M1",
                node_origin=df_node["node"],
                node_loc=df_node["node"],
            )
            .pipe(
                broadcast,
                yv_ya_sw,
                time=sub_time,
            )
            .pipe(same_time)
        )

        # input dataframe  for groundwater supply
        yv_ya_gw = map_yv_ya_lt(year_wat, 20, first_year)
        inp = inp.append(
            make_df(
                "input",
                technology="extract_groundwater",
                value=1,
                unit="-",
                level="water_avail_basin",
                commodity="groundwater_basin",
                mode="M1",
                node_origin=df_node["node"],
                node_loc=df_node["node"],
            )
            .pipe(
                broadcast,
                yv_ya_gw,
                time=sub_time,
            )
            .pipe(same_time)
        )

        # electricity input dataframe  for extract freshwater supply
        # low: 0.001141553, mid: 0.018835616, high: 0.03652968
        inp = inp.append(
            make_df(
                "input",
                technology="extract_surfacewater",
                value=0.018835616,
                unit="-",
                level="final",
                commodity="electr",
                mode="M1",
                time_origin="year",
                node_origin=df_node["region"],
                node_loc=df_node["node"],
            ).pipe(
                broadcast,
                yv_ya_sw,
                time=sub_time,
            )
        )

        inp = inp.append(
            make_df(
                "input",
                technology="extract_groundwater",
                value=df_gwt["GW_per_km3_per_year"] + 0.043464579,
                unit="-",
                level="final",
                commodity="electr",
                mode="M1",
                time_origin="year",
                node_origin=df_gwt["REGION"],
                node_loc=df_node["node"],
            ).pipe(
                broadcast,
                yv_ya_gw,
                time=sub_time,
            )
        )

        inp = inp.append(
            make_df(
                "input",
                technology="extract_gw_fossil",
                value=(df_gwt["GW_per_km3_per_year"] + 0.043464579)
                * 2,  # twice as much normal gw
                unit="-",
                level="final",
                commodity="electr",
                mode="M1",
                time_origin="year",
                node_origin=df_gwt["REGION"],
                node_loc=df_node["node"],
            ).pipe(
                broadcast,
                yv_ya_gw,
                time=sub_time,
            )
        )

        if context.type_reg == "global":
            inp.loc[
                (inp["technology"].str.contains("extract_gw_fossil"))
                & (inp["year_act"] == 2020)
                & (inp["node_loc"] == "R11_SAS"),
                "value",
            ] *= 0.5

        results["input"] = inp

        # Add output df  for freshwater supply for basins
        output_df = (
            make_df(
                "output",
                technology="extract_surfacewater",
                value=1,
                unit="-",
                level="water_supply_basin",
                commodity="freshwater_basin",
                mode="M1",
                node_loc=df_node["node"],
                node_dest=df_node["node"],
            )
            .pipe(
                broadcast,
                yv_ya_sw,
                time=sub_time,
            )
            .pipe(same_time)
        )
        # Add output df  for groundwater supply for basins
        output_df = output_df.append(
            make_df(
                "output",
                technology="extract_groundwater",
                value=1,
                unit="-",
                level="water_supply_basin",
                commodity="freshwater_basin",
                mode="M1",
                node_loc=df_node["node"],
                node_dest=df_node["node"],
            )
            .pipe(
                broadcast,
                yv_ya_gw,
                time=sub_time,
            )
            .pipe(same_time)
        )

        # Add output df  for groundwater supply for basins
        output_df = output_df.append(
            make_df(
                "output",
                technology="extract_gw_fossil",
                value=1,
                unit="-",
                level="water_supply_basin",
                commodity="freshwater_basin",
                mode="M1",
                node_loc=df_node["node"],
                node_dest=df_node["node"],
                time_origin="year",
            )
            .pipe(
                broadcast,
                yv_ya_gw,
                time=sub_time,
            )
            .pipe(same_time)
        )

        # Add output of saline water supply for regions
        output_df = output_df.append(
            make_df(
                "output",
                technology="extract_salinewater",
                value=1,
                unit="km3",
                year_vtg=year_wat,
                year_act=year_wat,
                level="saline_supply",
                commodity="saline_ppl",
                mode="M1",
                time="year",
                time_dest="year",
                time_origin="year",
            )
            .pipe(broadcast, node_loc=node_region)
            .pipe(same_node)
        )

        hist_new_cap = make_df(
            "historical_new_capacity",
            node_loc=df_hist["BCU_name"],
            technology="extract_surfacewater",
            value=df_hist["hist_cap_sw_km3_year"] / 5,  # n period
            unit="km3/year",
            year_vtg=2015,
        )

        hist_new_cap = hist_new_cap.append(
            make_df(
                "historical_new_capacity",
                node_loc=df_hist["BCU_name"],
                technology="extract_groundwater",
                value=df_hist["hist_cap_gw_km3_year"] / 5,
                unit="km3/year",
                year_vtg=2015,
            )
        )

        results["historical_new_capacity"] = hist_new_cap

        # output data frame linking water supply to energy dummy technology
        output_df = output_df.append(
            make_df(
                "output",
                technology="basin_to_reg",
                value=1,
                unit="-",
                level="water_supply",
                commodity="freshwater",
                time_dest="year",
                node_loc=df_node["region"],
                node_dest=df_node["region"],
                mode=df_node["mode"],
            ).pipe(broadcast, year_vtg=year_wat, time=sub_time)
        )

        output_df["year_act"] = output_df["year_vtg"]

        results["output"] = output_df

        # dummy variable cost for dummy water to energy technology
        var = make_df(
            "var_cost",
            technology="basin_to_reg",
            mode=df_node["mode"],
            node_loc=df_node["region"],
            value=20,
            unit="-",
        ).pipe(broadcast, year_vtg=year_wat, time=sub_time)
        var["year_act"] = var["year_vtg"]
        # # Dummy cost for extract surface ewater to prioritize water sources
        # var = var.append(make_df(
        #     "var_cost",
        #     technology='extract_surfacewater',
        #     value= 0.0001,
        #     unit="USD/km3",
        #     mode="M1",
        #     time="year",
        #     ).pipe(broadcast, year_vtg=year_wat,
        #       year_act=year_wat, node_loc=df_node["node"]
        #        )
        #                  )
        # # Dummy cost for extract groundwater
        # var = var.append(make_df(
        #     "var_cost",
        #     technology='extract_groundwater',
        #     value= 0.001,
        #     unit="USD/km3",
        #     mode="M1",
        #     time="year",
        # ).pipe(broadcast, year_vtg=year_wat,
        #   year_act=year_wat, node_loc=df_node["node"])
        #                )
        results["var_cost"] = var

        # load the share of sw
        df_sw = map_basin_region_wat(context)

        share = make_df(
            "share_mode_up",
            shares="share_basin",
            technology="basin_to_reg",
            mode=df_sw["mode"],
            node_share=df_sw["MSGREG"],
            time=df_sw["time"],
            value=df_sw["share"],
            unit="%",
            year_act=df_sw["year"],
        )

        results["share_mode_up"] = share

        tl = (
            make_df(
                "technical_lifetime",
                technology="extract_surfacewater",
                value=50,
                unit="y",
            )
            .pipe(broadcast, year_vtg=year_wat, node_loc=df_node["node"])
            .pipe(same_node)
        )

        tl = tl.append(
            make_df(
                "technical_lifetime",
                technology="extract_groundwater",
                value=20,
                unit="y",
            )
            .pipe(broadcast, year_vtg=year_wat, node_loc=df_node["node"])
            .pipe(same_node)
        )

        tl = tl.append(
            make_df(
                "technical_lifetime",
                technology="extract_gw_fossil",
                value=20,
                unit="y",
            )
            .pipe(broadcast, year_vtg=year_wat, node_loc=df_node["node"])
            .pipe(same_node)
        )

        results["technical_lifetime"] = tl

        # Prepare dataframe for investments
        inv_cost = make_df(
            "inv_cost",
            technology="extract_surfacewater",
            value=155.57,
            unit="USD/km3",
        ).pipe(broadcast, year_vtg=year_wat, node_loc=df_node["node"])

        inv_cost = inv_cost.append(
            make_df(
                "inv_cost",
                technology="extract_groundwater",
                value=54.52,
                unit="USD/km3",
            ).pipe(broadcast, year_vtg=year_wat, node_loc=df_node["node"])
        )

        inv_cost = inv_cost.append(
            make_df(
                "inv_cost",
                technology="extract_gw_fossil",
                value=54.52 * 150,  # assume higher as normal GW
                unit="USD/km3",
            ).pipe(broadcast, year_vtg=year_wat, node_loc=df_node["node"])
        )

        results["inv_cost"] = inv_cost

        fix_cost = make_df(
            "fix_cost",
            technology="extract_gw_fossil",
            value=300,  # assumed
            unit="USD/km3",
        ).pipe(broadcast, yv_ya_gw, node_loc=df_node["node"])

        results["fix_cost"] = fix_cost

    return results


def add_e_flow(context):
    """Add environmental flows
    This function bounds the available water and allocates the environmental
    flows.Environmental flow bounds are calculated using Variable Monthly Flow
    (VMF) method. The VMF method is applied to wet and dry seasonal runoff
    values. These wet and dry seasonal values are then aggregated to annual
    values.Environmental flows in the model will be incorporated as bounds on
    'return_flow' technology. The lower bound on this technology will ensure
    that certain amount of water remain
    Parameters
    ----------
    context : .Context
    Returns
    -------
    data : dict of (str -> pandas.DataFrame)
        Keys are MESSAGE parameter names such as 'input', 'fix_cost'.
        Values are data frames ready for :meth:`~.Scenario.add_par`.
        Years in the data include the model horizon indicated by
        ``context["water build info"]``, plus the additional year 2010.
    """
    # define an empty dictionary
    results = {}

    info = context["water build info"]

    # Adding freshwater supply constraints
    # Reading data, the data is spatially and temprally aggregated from GHMs
    df_sw, df_gw = read_water_availability(context)

    # reading sample for assiging basins
    PATH = package_data_path(
        "water", "delineation", f"basins_by_region_simpl_{context.regions}.csv"
    )
    df_x = pd.read_csv(PATH)

    dmd_df = make_df(
        "demand",
        node="B" + df_sw["Region"].astype(str),
        commodity="surfacewater_basin",
        level="water_avail_basin",
        year=df_sw["year"],
        time=df_sw["time"],
        value=df_sw["value"],
        unit="km3/year",
    )
    dmd_df = dmd_df[dmd_df["year"] >= 2025].reset_index(drop=True)
    dmd_df["value"] = dmd_df["value"].apply(lambda x: x if x >= 0 else 0)

    if "year" in context.time:
        # Reading data, the data is spatially and temporally aggregated from GHMs
        path1 = package_data_path(
            "water",
            "availability",
            f"e-flow_{context.RCP}_{context.regions}.csv",
        )
        df_env = pd.read_csv(path1)
        df_env.drop(["Unnamed: 0"], axis=1, inplace=True)
        df_env.index = df_x["BCU_name"]
        df_env = df_env.stack().reset_index()
        df_env.columns = ["Region", "years", "value"]
        df_env.sort_values(["Region", "years", "value"], inplace=True)
        df_env.fillna(0, inplace=True)
        df_env.reset_index(drop=True, inplace=True)
        df_env["year"] = pd.DatetimeIndex(df_env["years"]).year
        df_env["time"] = "year"
        df_env2210 = df_env[df_env["year"] == 2100]
        df_env2210["year"] = 2110
        df_env = pd.concat([df_env, df_env2210])
        df_env = df_env[df_env["year"].isin(info.Y)]
    else:
        # Reading data, the data is spatially and temporally aggregated from GHMs
        path1 = package_data_path(
            "water",
            "availability",
            f"e-flow_5y_m_{context.RCP}_{context.regions}.csv",
        )
        df_env = pd.read_csv(path1)
        df_env.drop(["Unnamed: 0"], axis=1, inplace=True)
        new_cols = pd.to_datetime(df_env.columns, format="%Y/%m/%d")
        df_env.columns = new_cols
        df_env.index = df_x["BCU_name"]
        df_env = df_env.stack().reset_index()
        df_env.columns = ["Region", "years", "value"]
        df_env.sort_values(["Region", "years", "value"], inplace=True)
        df_env.fillna(0, inplace=True)
        df_env.reset_index(drop=True, inplace=True)
        df_env["year"] = pd.DatetimeIndex(df_env["years"]).year
        df_env["time"] = pd.DatetimeIndex(df_env["years"]).month
        df_env2210 = df_env[df_env["year"] == 2100]
        df_env2210["year"] = 2110
        df_env = pd.concat([df_env, df_env2210])
        df_env = df_env[df_env["year"].isin(info.Y)]

    # Return a processed dataframe for env flow calculations
    if context.SDG != "baseline":
        # dataframe to put constraints on env flows
        eflow_df = make_df(
            "bound_activity_lo",
            node_loc="B" + df_env["Region"],
            technology="return_flow",
            year_act=df_env["year"],
            mode="M1",
            time=df_env["time"],
            value=df_env["value"],
            unit="km3/year",
        )

        eflow_df["value"] = eflow_df["value"].apply(lambda x: x if x >= 0 else 0)
        eflow_df = eflow_df[eflow_df["year_act"] >= 2025].reset_index(drop=True)

        dmd_df.sort_values(by=["node", "year"], inplace=True)
        dmd_df.reset_index(drop=True, inplace=True)
        eflow_df.sort_values(by=["node_loc", "year_act"], inplace=True)
        eflow_df.reset_index(drop=True, inplace=True)

        eflow_df["value"] = np.where(
            eflow_df["value"] >= 0.7 * dmd_df["value"],
            0.7 * dmd_df["value"],
            eflow_df["value"],
        )

        results["bound_activity_lo"] = eflow_df

    return results



"""Utilities for using :doc:`Pooch <pooch:about>`."""

import logging
from pathlib import Path
from typing import Tuple

import click
import pooch

from .context import Context

log = logging.getLogger(__name__)


class Extract:
    """Similar to :class:`pooch.Unzip`, streamlined using :mod:`pathlib`.

    This version supports:

    - Absolute or relative paths for the `extract_dir` parameter.
    - :file:`.zip` or :file:`.tar.xz` archives.
    """

    def __init__(self, members=None, extract_dir=None):
        self.members = members
        self.extract_dir = Path(extract_dir or ".")

    def __call__(self, fname, action, pooch):
        import tarfile
        import zipfile

        path = Path(fname)

        # Identify the directory for extracted files
        if self.extract_dir.is_absolute():
            # Some absolute path
            extract_dir = self.extract_dir
        else:
            # A relative path, possibly the default
            extract_dir = path.parent.joinpath(self.extract_dir)

        # Ensure the directory exists
        extract_dir.mkdir(parents=True, exist_ok=True)

        members = self.members

        # Select the class/method to open the archive, and the method name for listing
        # members
        cls, list_method = {
            ".zip": (zipfile.ZipFile, "namelist"),
            ".xz": (tarfile.TarFile.open, "getnames"),
        }[path.suffix]

        with cls(path) as archive:
            if members is None:
                members = getattr(archive, list_method)()
                log.info(f"Unpack all {len(members)} members of {path}")

            archive.extractall(members=members, path=extract_dir)

        return members


class UnpackSnapshot:
    """Pooch processor that calls :func:`.snapshot.unpack`."""

    def __call__(self, fname, action, pooch):
        from message_ix_models.model.snapshot import unpack

        path = Path(fname)
        unpack(path)

        return path


#: Supported remote sources of data.
SOURCE = {
    "PRIMAP": dict(
        pooch_args=dict(
            base_url="ftp://datapub.gfz-potsdam.de/download/10.5880.PIK.2019.001/",
            registry={
                "PRIMAP-hist_v2.0_11-Dec-2018.zip": (
                    "md5:f28d58abef4ecfc36fc8ce3e9eef2871"
                ),
            },
        ),
        processor=Extract(members=["PRIMAP-hist_v2.0_11-Dec-2018.csv"]),
    ),
    "MESSAGEix-Nexus": dict(
        pooch_args=dict(
            base_url="https://github.com/iiasa/message-ix-models/raw/enh/2023-W44/"
            "message_ix_models/data/water/",
            registry={"water.tar.xz": "sha1:ec9e0655af90ca844c0158968bb03a194b8fa6c6"},
        ),
        processor=Extract(extract_dir="water"),
    ),
    "snapshot-0": dict(
        pooch_args=dict(
            base_url="doi:10.5281/zenodo.5793870",
            registry={
                "MESSAGEix-GLOBIOM_1.1_R11_no-policy_baseline.xlsx": (
                    "md5:222193405c25c3c29cc21cbae5e035f4"
                ),
            },
        ),
        processor=UnpackSnapshot(),
    ),
}


def fetch(pooch_args, **fetch_kwargs) -> Tuple[Path, ...]:
    """Create a :class:`~pooch.Pooch` instance and fetch a single file.

    Files are stored under the directory identified by :meth:`.Context.get_cache_path`,
    unless `args` provides another location.

    Parameters
    ----------
    args
        Passed to :func:`pooch.create`.
    kwargs
        Passed to :meth:`pooch.Pooch.fetch`.

    Returns
    -------
    Path
        Path to the fetched file.

    See also
    --------
    :func:`.snapshot.load`
    """
    pooch_args.setdefault("path", Context.get_instance(-1).get_cache_path())

    p = pooch.create(**pooch_args)

    if len(p.registry) > 1:  # pragma: no cover
        raise NotImplementedError("fetch() with registries with >1 files")

    filenames = p.fetch(next(iter(p.registry.keys())), **fetch_kwargs)

    if isinstance(filenames, (str, Path)):
        filenames = [filenames]

    # Convert to pathlib.Path
    paths = tuple(map(Path, filenames))

    log.info(
        "Fetched"
        + (f" {paths[0]}" if len(paths) == 1 else "\n".join(map(str, (":",) + paths)))
    )

    return paths


@click.command("fetch")
@click.argument("source", metavar="SOURCE", type=click.Choice(list(SOURCE.keys())))
@click.pass_obj
def cli(context, source):
    """Retrieve data from primary sources."""
    from message_ix_models.util import pooch

    pooch.fetch(**SOURCE[source], progressbar=True)



import click

from message_ix_models.util.click import common_params


@click.group("res")
@click.pass_obj
def cli(context):
    """MESSAGEix-GLOBIOM reference energy system (RES)."""


@cli.command("create-bare")
@common_params("nodes")
@click.pass_obj
def create_bare(context):
    """Create the RES from scratch."""
    from .bare import create_res

    create_res(context)



"""Plots for MESSAGEix-GLOBIOM reporting.

The current set functions on time series data stored on the scenario by
:mod:`message_ix_models.report` or :mod:`message_data` legacy reporting.
"""

import logging
import re
from datetime import datetime
from typing import TYPE_CHECKING, List, Optional, Sequence

import genno.compat.plotnine
import pandas as pd
import plotnine as p9
from genno import Computer, Key

if TYPE_CHECKING:
    import plotnine.typing
    from genno.core.key import KeyLike
    from message_ix import Scenario

    from message_ix_models import Context

__all__ = [
    "PLOTS",
    "EmissionsCO2",
    "FinalEnergy0",
    "FinalEnergy1",
    "Plot",
    "PrimaryEnergy0",
    "PrimaryEnergy1",
    "callback",
]

log = logging.getLogger(__name__)


class Plot(genno.compat.plotnine.Plot):
    """Base class for plots based on reported time-series data.

    Subclasses should be used like:

    .. code-block:: python

       class MyPlot(Plot):
           ...

       c.add("plot myplot", MyPlot, "scenario")

    …that is, giving "scenario" or another key that points to a |Scenario| object with
    stored time series data. See the examples in this file.
    """

    #: 'Static' geoms: list of plotnine objects that are not dynamic.
    static: List["plotnine.typing.PlotAddable"] = [
        p9.theme(figure_size=(23.4, 16.5)),  # A3 paper in landscape [inches]
        # p9.theme(figure_size=(11.7, 8.3)),  # A4 paper in landscape
    ]

    #: Fixed plot title string. If not given, the first line of the class docstring is
    #: used.
    title = None

    #: Units expression for plot title.
    unit = None

    #: Scenario URL for plot title.
    url: Optional[str] = None

    # NB only here to narrow typing
    inputs: Sequence[str] = []

    #: List of regular expressions corresponding to :attr:`inputs`. These are passed as
    #: the `expr` argument to :func:`filter_ts` to filter the entire set of time series
    #: data.
    inputs_regex: List[re.Pattern] = []

    @classmethod
    def add_tasks(
        cls, c: "Computer", key: "KeyLike", *inputs, strict: bool = False
    ) -> "KeyLike":
        from copy import copy
        from itertools import zip_longest

        scenario_key = inputs[0]

        # Retrieve all time series data, for advanced filtering
        all_data = Key(scenario_key) + "iamc"
        c.add(all_data, "get_ts", scenario_key)

        if len(cls.inputs_regex):
            # Iterate over matched items from `inputs` and `inputs_regex`
            for k, expr in zip_longest(cls.inputs, cls.inputs_regex):
                if expr is None:
                    break
                # Filter the data given by `expr` from all::iamc
                c.add(k, "filter_ts", all_data, copy(expr))
        else:
            for k in map(Key, cls.inputs):
                # Add a computation to get the time series data for a specific variable
                c.add(k, "get_ts", scenario_key, dict(variable=k.name))

        # Add the plot itself
        return super().add_tasks(c, key, *inputs[1:], strict=strict)

    def ggtitle(self, value=None) -> p9.ggtitle:
        """Return :class:`plotnine.ggtitle` including the current date & time."""
        title_pieces = [
            (self.title or self.__doc__ or "").splitlines()[0].rstrip("."),
            f"[{self.unit}]" if self.unit else None,
            value,
            "\n",
            self.url,
            f"({datetime.now().isoformat(timespec='minutes')})",
        ]
        return p9.ggtitle(" ".join(filter(None, title_pieces)))

    def groupby_plot(self, data: pd.DataFrame, *args):
        """Combination of groupby and ggplot().

        Groups by `args` and yields a series of :class:`plotnine.ggplot` objects, one
        per group, with :attr:`static` geoms and :func:`ggtitle` appended to each.
        """
        for group_key, group_df in data.groupby(*args):
            yield (
                group_key,
                (
                    p9.ggplot(group_df)
                    + self.static
                    + self.ggtitle(
                        group_key if isinstance(group_key, str) else repr(group_key)
                    )
                ),
            )


class EmissionsCO2(Plot):
    """CO₂ Emissions."""

    basename = "emission-CO2"
    inputs = ["Emissions|CO2::iamc", "scenario"]

    static = Plot.static + [
        p9.aes(x="year", y="value", color="region"),
        p9.geom_line(),
        p9.geom_point(),
        p9.labs(x="Period", y="", color="Region"),
    ]

    def generate(self, data: pd.DataFrame, scenario: "Scenario"):
        self.url = scenario.url
        self.unit = data["unit"].unique()[0]

        for _, ggplot in self.groupby_plot(data, data.region.str.contains("GLB")):
            y_max = max(ggplot.data["value"])
            yield ggplot + p9.expand_limits(y=[0, y_max]) + self.ggtitle("")


class FinalEnergy0(EmissionsCO2):
    """Final Energy."""

    basename = "fe0"
    inputs = ["Final Energy::iamc", "scenario"]


class FinalEnergy1(Plot):
    """Final Energy."""

    basename = "fe1"
    inputs = ["fe1-0::iamc", "scenario"]

    _c = [
        "Electricity",
        "Gases",
        "Geothermal",
        "Heat",
        "Hydrogen",
        "Liquids",
        "Solar",
        "Solids",
    ]
    inputs_regex = [re.compile(rf"Final Energy\|({'|'.join(_c)})")]

    static = Plot.static + [
        p9.aes(x="year", y="value", fill="variable"),
        p9.geom_bar(stat="identity", size=5.0),  # 5.0 is the minimum spacing of "year"
        p9.labs(x="Period", y="", fill="Commodity"),
    ]

    def generate(self, data: pd.DataFrame, scenario: "Scenario"):
        self.url = scenario.url
        self.unit = data["unit"].unique()[0]

        for _, ggplot in self.groupby_plot(data, "region"):
            yield ggplot


class PrimaryEnergy0(EmissionsCO2):
    """Primary Energy."""

    basename = "pe0"
    inputs = ["Primary Energy::iamc", "scenario"]


class PrimaryEnergy1(FinalEnergy1):
    """Primary Energy."""

    basename = "pe1"
    inputs = ["pe1-0::iamc", "scenario"]

    _omit = ["Fossil", "Non-Biomass Renewables", "Secondary Energy Trade"]
    inputs_regex = [re.compile(rf"Primary Energy\|((?!{'|'.join(_omit)})[^\|]*)")]


#: All plot classes.
PLOTS = (
    EmissionsCO2,
    FinalEnergy0,
    FinalEnergy1,
    PrimaryEnergy0,
    PrimaryEnergy1,
)


def callback(c: Computer, context: "Context") -> None:
    """Add all :data:`PLOTS` to `c`.

    Also add a key "plot all" to triggers the generation of all plots.
    """
    all_keys = [c.add(f"plot {p.basename}", p, "scenario") for p in PLOTS]
    c.add("plot all", all_keys)
    log.info(f"Add 'plot all' collecting {len(all_keys)} plots")



from message_ix_models.project.advance.data import LOCATION, NAME
from message_ix_models.tools.iamc import describe
from message_ix_models.util import MESSAGE_MODELS_PATH, package_data_path


def test_describe(test_context):
    import zipfile

    import pandas as pd

    path = package_data_path("test", *LOCATION)
    with zipfile.ZipFile(path) as zf:
        data = pd.read_csv(zf.open(NAME), engine="pyarrow").rename(
            columns=lambda c: c.upper()
        )

    sm = describe(
        data, f"ADVANCE data in {path.relative_to(MESSAGE_MODELS_PATH.parent)}"
    )

    # Message contains the expected code lists.
    # Code lists have the expected lengths.
    for id, N in (
        ("MODEL", 12),
        ("SCENARIO", 51),
        ("REGION", 14),
        ("VARIABLE", 3080),
        ("UNIT", 29),
    ):
        assert N == len(sm.codelist[id])

    # from message_ix_models.util.sdmx import write

    # write(sm, basename="ADVANCE")



"""Tools for calibrating MACRO for MESSAGEix-GLOBIOM.

See :doc:`message-ix:macro` for *general* documentation on MACRO and MESSAGE-MACRO. This
module contains tools specifically for using these models with MESSAGEix-GLOBIOM.
"""

import logging
from functools import lru_cache
from itertools import product
from pathlib import Path
from typing import TYPE_CHECKING, List, Literal, Mapping, Optional, Union

import pandas as pd

from message_ix_models.model.bare import get_spec
from message_ix_models.util import nodes_ex_world

if TYPE_CHECKING:
    from sdmx.model.v21 import Code

    from message_ix_models import Context

log = logging.getLogger(__name__)

#: Default set of commodities to include in :func:`generate`.
COMMODITY = ["i_therm", "i_spec", "rc_spec", "rc_therm", "transport"]


def generate(
    parameter: Literal["aeei", "config", "depr", "drate", "lotol"],
    context: "Context",
    commodities: Union[List[str], List["Code"]] = COMMODITY,
    value: Optional[float] = None,
) -> pd.DataFrame:
    """Generate uniform data for one :mod:`message_ix.macro` `parameter`.

    :meth:`message_ix.Scenario.add_macro` expects as its `data` parameter a
    :class:`dict` that maps certain MACRO parameter names (or the special name "config")
    to :class:`.pandas.DataFrame`. This function generates data for those data frames.

    For the particular dimensions, generate automatically includes:

    - "node": All nodes in the node code list given by :func:`.nodes_ex_world`, for the
      node list indicated by :attr:`.model.Config.regions`.
    - "year": All periods from the period *before* the first model year.
    - "commodity": The elements of `commodities`.
    - "sector": If each entry of `commodities` is a |Code| and has an annotation with
      id="macro-sector", the value of that annotation. Otherwise, the same as
      `commodity`.

    `value` supplies the parameter value, which is the same for all observations.
    The labels level="useful" and unit="-" are fixed.

    Parameters
    ----------
    parameter : str
        MACRO parameter for which to generate data.
    context
        Used with :func:`.bare.get_spec`.
    commodities : list of str or |Code|
        Commodities to include in the MESSAGE-MACRO linkage.
    value : float
        Parameter value.

    Returns
    -------
    pandas.DataFrame
        The columns vary according to `parameter`:

        - "aeei": node, sector, year, value, unit.
        - "depr", "drate", or "lotol": node, value, unit.
        - "config": node, sector, commodity, level, year.
    """
    spec = get_spec(context)

    if isinstance(commodities[0], str):
        c_codes = spec.add.set["commodity"]
    else:
        c_codes = commodities

    @lru_cache
    def _sector(commodity: str) -> str:
        try:
            idx = c_codes.index(commodity)
            return str(c_codes[idx].get_annotation(id="macro-sector").text)
        except (KeyError, ValueError) as e:
            log.info(e)
            return str(commodity)

    # AEEI data must begin from the period before the first model period
    y0_index = spec.add.set["year"].index(spec.add.y0)
    iterables = dict(
        c_s=zip(  # Paired commodity and sector
            map(str, commodities), map(_sector, commodities)
        ),
        level=["useful"],
        node=nodes_ex_world(spec.add.N),
        sector=map(_sector, commodities),
        year=spec.add.set["year"][y0_index:],
    )

    if parameter == "aeei":
        dims = ["node", "year", "sector"]
        iterables.update(year=spec.add.set["year"][y0_index - 1 :])
    elif parameter == "config":
        dims = ["node", "c_s", "level", "year"]
        assert value is None
    elif parameter in ("depr", "drate", "lotol"):
        dims = ["node"]
    else:
        raise NotImplementedError(f"generate(…) for MACRO parameter {parameter!r}")

    result = pd.DataFrame(
        [tuple(values) for values in product(*[iterables[d] for d in dims])],
        columns=dims,
    )

    if parameter == "config":
        return pd.concat(
            [
                result.drop("c_s", axis=1),
                pd.DataFrame(result["c_s"].tolist(), columns=["commodity", "sector"]),
            ],
            axis=1,
        )
    else:
        return result.assign(value=value, unit="-")


def load(base_path: Path) -> Mapping[str, pd.DataFrame]:
    """Load MACRO data from CSV files.

    The function reads files in the simple/long CSV format understood by
    :func:`.genno.operator.load_file`. For use with
    :meth:`~message_ix.Scenario.add_macro`, the dimension names should be given in full,
    for instance "node" or "sector".

    Parameters
    ----------
    base_path : pathlib.Path
        Directory containing zero or more CSV files.

    Returns
    -------
    dict of (str -> pandas.DataFrame)
        Mapping from MACRO calibration parameter names to data; one entry for each file
        in `base_path`.
    """
    from genno.operator import load_file

    result = {}
    for filename in base_path.glob("*.csv"):
        name = filename.stem

        q = load_file(filename, name=name)

        result[name] = (
            q.to_frame()
            .reset_index()
            .rename(columns={name: "value"})
            .assign(unit=f"{q.units:~}" or "-")
        )

    return result



import logging
import os
from dataclasses import dataclass, field, fields, is_dataclass, replace
from pathlib import Path
from typing import Any, Hashable, List, Mapping, MutableMapping, Optional, Sequence, Set

import ixmp

from .scenarioinfo import ScenarioInfo

log = logging.getLogger(__name__)

ixmp.config.register("no message_data", bool, False)
ixmp.config.register("message local data", Path, Path.cwd())


def _local_data_factory():
    """Default values for :attr:`.Config.local_data."""
    return (
        Path(
            os.environ.get("MESSAGE_LOCAL_DATA", None)
            or ixmp.config.get("message local data")
        )
        .expanduser()
        .resolve()
    )


@dataclass
class ConfigHelper:
    """Mix-in for :class:`dataclass`-based configuration classes.

    This provides 3 methods—:meth:`read_file`, :meth:`replace`, and :meth:`from_dict`—
    that help to use :class:`dataclass` classes for handling :mod:`message_ix_models`
    configuration.

    All 3 methods take advantage of name manipulations: the characters "-" and " " are
    replaced with underscores ("_"). This allows to write the names of attributes in
    legible ways—e.g. "attribute name" or “attribute-name” instead of "attribute_name"—
    in configuration files and/or code.
    """

    @classmethod
    def _fields(cls) -> Set[str]:
        """Names of fields in `cls`."""
        result = set(dir(cls))
        if is_dataclass(cls):
            result |= set(map(lambda f: f.name, fields(cls)))
        return result

    @classmethod
    def _canonical_name(cls, name: Hashable) -> Optional[str]:
        """Canonicalize a name into a valid Python attribute name."""
        _name = str(name).replace(" ", "_").replace("-", "_")
        return _name if _name in cls._fields() else None

    @classmethod
    def _munge_dict(cls, data: Mapping[Hashable, Any], fail: str, kind: str):
        for key, value in data.items():
            name = cls._canonical_name(key)

            if name:
                yield name, value
            else:
                msg = f"{cls.__name__} has no attribute for {kind} {key!r}"
                if fail == "raise":
                    raise ValueError(msg)
                else:
                    log.info(f"{msg}; ignored")

    def read_file(self, path: Path, fail="raise") -> None:
        """Update configuration from file.

        Parameters
        ----------
        path
            to a :file:`.yaml` file containing a top-level mapping.
        fail : str
            if "raise" (the default), any names in `path` which do not match attributes
            of the dataclass raise a ValueError. Ottherwise, a message is logged.
        """
        if path.suffix == ".yaml":
            import yaml

            with open(path, encoding="utf-8") as f:
                data = yaml.safe_load(f)
        elif path.suffix == ".json":
            import json

            with open(path) as f:
                data = json.load(f)
        else:
            raise NotImplementedError(f"Read from {path.suffix}")

        for key, value in self._munge_dict(data, fail, "file section"):
            existing = getattr(self, key, None)
            if is_dataclass(existing) and not isinstance(existing, type):
                # Attribute value is also a dataclass; update it recursively
                if isinstance(existing, ConfigHelper):
                    # Use name manipulation on the attribute value also
                    value = existing.replace(**value)
                elif not isinstance(existing, type):
                    # https://github.com/python/mypy/issues/15843
                    # TODO Check that fix is available in mypy 1.7.x; remove
                    value = replace(existing, **value)  # type: ignore [misc]
            setattr(self, key, value)

    def replace(self, **kwargs):
        """Like :func:`dataclasses.replace` with name manipulation."""
        return replace(
            self,
            **{k: v for k, v in self._munge_dict(kwargs, "raise", "keyword argument")},
        )

    def update(self, **kwargs):
        """Update attributes in-place.

        Raises
        ------
        AttributeError
            Any of the `kwargs` are not fields in the data class.
        """
        # TODO use _munge_dict(); allow a positional argument
        for k, v in kwargs.items():
            if not hasattr(self, k):
                raise AttributeError(k)
            setattr(self, k, v)

    @classmethod
    def from_dict(cls, data: Mapping):
        """Construct an instance from `data` with name manipulation."""
        return cls(**{k: v for k, v in cls._munge_dict(data, "raise", "mapping key")})


@dataclass
class Config:
    """Top-level configuration for :mod:`message_ix_models` and :mod:`message_data`."""

    #: Base path for :ref:`system-specific data <local-data>`, i.e. as given by the
    #: :program:`--local-data` CLI option or `message local data` key in the ixmp
    #: configuration file.
    local_data: Path = field(default_factory=_local_data_factory)

    #: Keyword arguments—especially `name`—for the :class:`ixmp.Platform` constructor,
    #: from the :program:`--platform` or :program:`--url` CLI option.
    platform_info: MutableMapping[str, str] = field(default_factory=dict)

    #: Keyword arguments—`model`, `scenario`, and optionally `version`—for the
    #: :class:`ixmp.Scenario` constructor, as given by the :program:`--model`/
    #: :program:`--scenario` or :program:`--url` CLI options.
    scenario_info: MutableMapping[str, str] = field(default_factory=dict)

    #: Like `scenario_info`, but a list for operations affecting multiple scenarios.
    scenarios: List[ScenarioInfo] = field(default_factory=list)

    #: Like :attr:`platform_info`, used by e.g. :meth:`.clone_to_dest`.
    dest_platform: MutableMapping[str, str] = field(default_factory=dict)

    #: Like :attr:`scenario_info`, used by e.g. :meth:`.clone_to_dest`.
    dest_scenario: MutableMapping[str, str] = field(default_factory=dict)

    #: A scenario URL, e.g. as given by the :program:`--url` CLI option.
    url: Optional[str] = None

    #: Like :attr:`url`, used by e.g. :meth:`.clone_to_dest`.
    dest: Optional[str] = None

    #: Base path for cached data, e.g. as given by the :program:`--cache-path` CLI
    #: option. Default: the directory :file:`message-ix-models` within the directory
    #: given by :func:`.platformdirs.user_cache_path`.
    cache_path: Optional[str] = None

    #: Paths of files containing debug outputs. See :meth:`Context.write_debug_archive`.
    debug_paths: Sequence[str] = field(default_factory=list)

    #: Whether an operation should be carried out, or only previewed. Different modules
    #: will respect :attr:`dry_run` in distinct ways, if at all, and **should** document
    #: behaviour.
    dry_run: bool = False

    #: Flag for causing verbose output to logs or stdout. Different modules will respect
    #: :attr:`verbose` in distinct ways.
    verbose: bool = False

    def __post_init__(self):
        if self.cache_path is None:
            from platformdirs import user_cache_path

            self.cache_path = user_cache_path("message-ix-models", ensure_exists=True)



import pytest
from genno import Computer

from message_ix_models.tools.exo_data import (
    DemoSource,
    ExoDataSource,
    prepare_computer,
    register_source,
)


class TestExoDataSource:
    def test_abstract(self):
        with pytest.raises(TypeError, match="Can't instantiate"):
            ExoDataSource()

    def test_register_source(self):
        with pytest.raises(ValueError, match="already registered for"):
            register_source(DemoSource)


@pytest.mark.parametrize("regions, N_n", [("R12", 12), ("R14", 14)])
def test_prepare_computer(test_context, regions, N_n):
    """:func:`.exo_data.prepare_computer` works as intended."""
    test_context.model.regions = regions

    c = Computer()

    # Function runs successfully `c`
    keys = prepare_computer(test_context, c, "test s1", dict(measure="POP"))

    # print(c.describe(keys[-1]))

    # Computation of data runs successfully
    result = c.get(keys[-1])

    # Data has the expected dimensions
    assert ("n", "y") == result.dims

    # Data is complete
    assert N_n == len(result.coords["n"])
    assert 14 == len(result.coords["y"])


def test_prepare_computer_exc(test_context):
    c = Computer()

    with pytest.raises(ValueError, match="No source found that can handle"):
        prepare_computer(test_context, c, "test s1", dict(measure="FOO"))

    with pytest.raises(ValueError, match="No source found that can handle"):
        prepare_computer(test_context, c, "not a source")


@pytest.mark.parametrize("regions, N_n", [("R12", 12), ("R14", 14)])
def test_operator(test_context, regions, N_n):
    """Exogenous data calculations can be set up through :meth:`.Computer.add`."""
    test_context.model.regions = regions

    c = Computer()
    c.require_compat("message_ix_models.report.operator")

    # Function runs successfully `c`
    keys = c.add(
        "exogenous_data",
        context=test_context,
        source="test s1",
        source_kw=dict(measure="POP"),
    )

    # print(c.describe(keys[-1]))

    # Computation of data runs successfully
    result = c.get(keys[-1])

    # Data has the expected dimensions
    assert ("n", "y") == result.dims

    # Data is complete
    assert N_n == len(result.coords["n"])
    assert 14 == len(result.coords["y"])



import logging
from typing import Callable, Dict, List, Mapping, Optional, Union

import ixmp
import pandas as pd
from message_ix import Scenario
from sdmx.model.v21 import Code

from message_ix_models.util import add_par_data, strip_par_data
from message_ix_models.util.ixmp import maybe_check_out, maybe_commit
from message_ix_models.util.scenarioinfo import ScenarioInfo, Spec

log = logging.getLogger(__name__)


def _add_unit(mp: ixmp.Platform, unit: str, comment: str) -> None:
    """Handle exceptions in :meth:`.Platform.add_unit`."""
    # TODO move upstream to ixmp.JDBCBackend
    log.info(f"Add unit {repr(unit)}")
    try:
        mp.add_unit(unit, comment)
    except Exception as e:  # pragma: no cover
        if "Error assigning an unit-key-id mapping" in str(e) and "" == str(unit):
            log.warning(f"…skip {repr(unit)} (ixmp.JDBCBackend with Oracle database)")
        else:
            raise


# FIXME Reduce complexity from 14 to ≤13
def apply_spec(  # noqa: C901
    scenario: Scenario,
    spec: Union[Spec, Mapping[str, ScenarioInfo]],
    data: Optional[Callable] = None,
    **options,
):
    """Apply `spec` to `scenario`.

    Parameters
    ----------
    spec : .Spec
        Specification of changes to make to `scenario`.
    data : callable, *optional*
        Function to add data to `scenario`. `data` can either manipulate the scenario
        directly, or return a :class:`dict` compatible with :func:`.add_par_data`.

    Other parameters
    ----------------
    dry_run : bool
        Don't modify `scenario`; only show what would be done. Default :obj:`False`.
        Exceptions will still be raised if the elements from ``spec['required']`` are
        missing; this serves as a check that the scenario has the required features for
        applying the spec.
    fast : bool
        Do not remove existing parameter data; increases speed on large scenarios.
    quiet : bool
        Only show log messages at level ``ERROR`` and higher. If :obj:`False` (default),
        show log messages at level ``DEBUG`` and higher.
    message : str
        Commit message.

    See also
    --------
    .add_par_data
    .strip_par_data
    .Code
    .ScenarioInfo
    """
    dry_run = options.get("dry_run", False)
    fast = options.get("fast", False)

    log.setLevel(logging.ERROR if options.get("quiet", False) else logging.DEBUG)

    if not dry_run:
        try:
            scenario.remove_solution()
        except ValueError:
            pass
        maybe_check_out(scenario)

    dump: Dict[str, pd.DataFrame] = {}  # Removed data

    # Sort the list of sets by the number of dimensions; this places basic (non-indexed)
    # sets first. Elements for these sets must be added before elements for indexed
    # sets that may reference them.
    sets = sorted((len(scenario.idx_sets(s)), s) for s in scenario.set_list())

    for _, set_name in sets:
        # Check whether this set is mentioned at all in the spec
        if 0 == sum(map(lambda info: len(info.set[set_name]), spec.values())):
            # Not mentioned; don't do anything
            continue

        log.info(f"Set {repr(set_name)}")

        # Base contents of the set
        base_set = scenario.set(set_name)
        # Unpack a multi-dimensional/indexed set to a list of tuples
        base = (
            list(base_set.itertuples(index=False))
            if isinstance(base_set, pd.DataFrame)
            else base_set.tolist()
        )

        log.info(f"  {len(base)} elements")
        # log.debug(', '.join(map(repr, base)))  # All elements; verbose

        # Check for required elements
        require = spec["require"].set[set_name]
        log.info(f"  Check {len(require)} required elements")

        # Raise an exception about the first missing element
        missing = list(filter(lambda e: e not in base, require))
        if missing:
            log.error(f"  {len(missing)} elements not found: {missing!r}")
            raise ValueError

        # Remove elements and associated parameter values
        for element in spec["remove"].set[set_name]:
            strip_par_data(
                scenario,
                set_name,
                element,
                dry_run=dry_run,
                dump=None if fast else dump,
            )

        # Add elements
        add = [] if dry_run else spec["add"].set[set_name]
        for element in add:
            name = element.id if isinstance(element, Code) else element
            scenario.add_set(set_name, name)
            if set_name == "node":
                scenario.platform.add_region(name, "region")

        if len(add):
            log.info(f"  Add {len(add)} element(s)")
            log.debug("  " + ellipsize(add))

        log.info("  ---")

    if not fast:
        N_removed = sum(len(d) for d in dump.values())
        log.info(f"{N_removed} total rows removed")

    # Add units to the Platform before adding data
    for unit in spec["add"].set["unit"]:
        unit = unit if isinstance(unit, Code) else Code(id=unit, name=unit)
        _add_unit(scenario.platform, unit.id, str(unit.name))

    # Add data
    if callable(data):
        result = data(scenario, dry_run=dry_run)
        if result:
            # `data` function returned some data; use add_par_data()
            add_par_data(scenario, result, dry_run=dry_run)

    # Finalize
    log.info("Commit results.")
    maybe_commit(
        scenario,
        condition=not dry_run,
        message=options.get("message", f"{__name__}.apply_spec()"),
    )


def ellipsize(elements: List) -> str:
    """Generate a short string representation of `elements`.

    If the list has more than 5 elements, only the first two and last two are shown,
    with "..." between.
    """
    if len(elements) > 5:
        return ", ".join(map(str, elements[:2] + ["..."] + elements[-2:]))
    else:
        return ", ".join(map(str, elements))



"""Tests of :mod:`.model.disutility`."""

from itertools import product

import pandas as pd
import pandas.testing as pdt
import pytest
from message_ix import make_df
from sdmx.model.v21 import Annotation, Code

from message_ix_models import ScenarioInfo, testing
from message_ix_models.model import disutility
from message_ix_models.util import (
    add_par_data,
    copy_column,
    make_source_tech,
    merge_data,
)

# Common data and fixtures for test_minimal() and other tests

COMMON = dict(
    level="useful",
    mode="all",
    node_dest="R14_AFR",
    node_loc="R14_AFR",
    node_origin="R14_AFR",
    node="R14_AFR",
    time_dest="year",
    time_origin="year",
    time="year",
    unit="kg",
)


@pytest.fixture
def groups():
    """Fixture: list of 2 consumer groups."""
    yield [Code(id="g0"), Code(id="g1")]


@pytest.fixture
def techs():
    """Fixture: list of 2 technologies for which groups can have disutility."""
    yield [Code(id="t0"), Code(id="t1")]


@pytest.fixture
def template():
    """Fixture: :class:.`Code` with annotations, for :func:`.disutility.get_spec`."""
    # Template for inputs of conversion technologies, from a technology-specific
    # commodity
    input = dict(commodity="output of {technology}", level="useful", unit="kg")

    # Template for outputs of conversion technologies, to a group–specific demand
    # commodity
    output = dict(commodity="demand of group {group}", level="useful", unit="kg")

    # Code's ID is itself a template for IDs of conversion technologies
    yield Code(
        id="usage of {technology} by {group}",
        annotations=[
            Annotation(id="input", text=repr(input)),
            Annotation(id="output", text=repr(output)),
        ],
    )


@pytest.fixture
def spec(groups, techs, template):
    """Fixture: a prepared spec for the minimal test case."""
    yield disutility.get_spec(groups, techs, template)


@pytest.fixture
def scenario(request, test_context, techs):
    """Fixture: a |Scenario| with technologies given by :func:`techs`."""
    test_context.regions = "R14"
    s = testing.bare_res(request, test_context, solved=False)
    s.check_out()

    s.add_set("technology", ["t0", "t1"])

    s.commit("Test fixture for .model.disutility")
    yield s


def test_add(scenario, groups, techs, template):
    """:func:`.disutility.add` runs on the bare RES; the result solves."""
    disutility.add(scenario, groups, techs, template)

    # Scenario solves (no demand)
    scenario.solve(quiet=True)
    assert (scenario.var("ACT")["lvl"] == 0).all()


def minimal_test_data(scenario):
    """Generate data for :func:`test_minimal`."""
    common = COMMON.copy()
    common.pop("node_loc")
    common.update(dict(mode="all"))

    data = dict()

    info = ScenarioInfo(scenario)
    y0 = info.Y[0]
    y1 = info.Y[1]

    # Output from t0 and t1
    for t in ("t0", "t1"):
        common.update(dict(technology=t, commodity=f"output of {t}"))
        merge_data(data, make_source_tech(info, common, output=1.0, var_cost=1.0))

    # Disutility input for each combination of (tech) × (group) × (2 years)
    input_data = pd.DataFrame(
        [
            ["usage of t0 by g0", y0, 0.1],
            ["usage of t0 by g0", y1, 0.1],
            ["usage of t1 by g0", y0, 0.1],
            ["usage of t1 by g0", y1, 0.1],
            ["usage of t0 by g1", y0, 0.1],
            ["usage of t0 by g1", y1, 0.1],
            ["usage of t1 by g1", y0, 0.1],
            ["usage of t1 by g1", y1, 0.1],
        ],
        columns=["technology", "year_vtg", "value"],
    )
    data["input"] = make_df(
        "input", **input_data, commodity="disutility", **COMMON
    ).assign(node_origin=copy_column("node_loc"), year_act=copy_column("year_vtg"))

    # Demand
    c, y = zip(*product(["demand of group g0", "demand of group g1"], [y0, y1]))
    data["demand"] = make_df("demand", commodity=c, year=y, value=1.0, **COMMON)

    # Constraint on activity in the first period
    t = sorted(input_data["technology"].unique())
    for bound in ("lo", "up"):
        par = f"bound_activity_{bound}"
        data[par] = make_df(par, value=0.5, technology=t, year_act=y0, **COMMON)

    # Constraint on activity growth
    annual = (1.1 ** (1.0 / 5.0)) - 1.0
    for bound, factor in (("lo", -1.0), ("up", 1.0)):
        par = f"growth_activity_{bound}"
        data[par] = make_df(
            par, value=factor * annual, technology=t, year_act=y1, **COMMON
        )

    return data, y0, y1


def test_minimal(scenario, groups, techs, template):
    """Expected results are generated from a minimal test case."""
    # Set up structure
    disutility.add(scenario, groups, techs, template)

    # Add test-specific data
    data, y0, y1 = minimal_test_data(scenario)

    scenario.check_out()
    add_par_data(scenario, data)
    scenario.commit("Disutility test 1")

    # commented: pre-solve debugging output
    # for par in ("input", "output", "technical_lifetime", "var_cost"):
    #     scenario.par(par).to_csv(f"debug-{par}.csv")

    scenario.solve(quiet=True)

    # Helper function to retrieve ACT data and condense for inspection
    def get_act(s):
        result = (
            scenario.var("ACT")
            .query("lvl > 0")
            .drop(columns=["node_loc", "mode", "time", "mrg"])
            .sort_values(["year_vtg", "technology"])
            .reset_index(drop=True)
        )
        # No "stray" activity of technologies beyond the vintage periods
        pdt.assert_series_equal(
            result["year_act"], result["year_vtg"], check_names=False
        )
        result = result.drop(columns=["year_vtg"]).set_index(["technology", "year_act"])
        # Return the activity and its inter-period delta
        return result, (
            result.xs(y1, level="year_act") - result.xs(y0, level="year_act")
        )

    # Post-solve debugging output TODO comment before merging
    ACT1, ACT1_delta = get_act(scenario)

    # Increase the disutility of for t0 for g0 in period y1
    data["input"].loc[1, "value"] = 0.2

    # Re-solve
    scenario.remove_solution()
    scenario.check_out()
    scenario.add_par("input", data["input"])
    scenario.commit("Disutility test 2")
    scenario.solve(quiet=True)

    # Compare activity
    ACT2, ACT2_delta = get_act(scenario)

    merged = ACT1.merge(ACT2, left_index=True, right_index=True)
    merged["lvl_diff"] = merged["lvl_y"] - merged["lvl_x"]

    merged_delta = ACT1_delta.merge(ACT2_delta, left_index=True, right_index=True)

    # commented: for debugging
    # print(merged, merged_delta)

    # Group g0 decreases usage of t0, and increases usage of t1, in period y1 vs. y0
    assert merged_delta.loc["usage of t0 by g0", "lvl_y"] < 0
    assert merged_delta.loc["usage of t1 by g0", "lvl_y"] > 0

    # Group g0 usage of t0 is lower when the disutility is higher
    assert merged.loc[("usage of t0 by g0", y1), "lvl_diff"] < 0
    # Group g0 usage of t1 is correspondingly higher
    assert merged.loc[("usage of t1 by g0", y1), "lvl_diff"] > 0


def test_data_conversion(scenario, spec):
    """:func:`~.disutility.data_conversion` runs."""
    info = ScenarioInfo(scenario)
    disutility.data_conversion(info, spec)


def test_data_source(scenario, spec):
    """:func:`~.disutility.data_source` runs."""
    info = ScenarioInfo(scenario)
    disutility.data_source(info, spec)


def test_get_data(scenario, spec):
    """:func:`~.disutility.get_data` runs."""
    data = disutility.get_data(scenario, spec)

    # Test that the code will not encounter #45 / iiasa/ixmp#425
    for name, df in data.items():
        assert (
            "" not in df["unit"].unique()
        ), f"{repr(name)} data has dimensionless units"


def test_get_spec(groups, techs, template):
    """:func:`~.disutility.get_spec` runs and produces expected output."""
    spec = disutility.get_spec(groups, techs, template)

    # Spec requires the existence of the base technologies
    assert {"technology"} == set(spec["require"].set.keys())
    assert techs == spec["require"].set["technology"]

    # Spec removes nothing
    assert set() == set(spec["remove"].set.keys())

    # Spec adds the "disutility" commodity; and adds (if not existing) the output
    # commodities for t[01] and demand commodities for g[01]
    assert {
        "disutility",
        "output of t0",
        "output of t1",
        "demand of group g0",
        "demand of group g1",
    } == set(map(str, spec["add"].set["commodity"]))

    # Spec adds the "distuility source" technology, and "usage of {tech} by {group}"
    # for each tech × group, per the template
    assert {
        "disutility source",
        "usage of t0 by g0",
        "usage of t0 by g1",
        "usage of t1 by g0",
        "usage of t1 by g1",
    } == set(map(str, spec["add"].set["technology"]))



"""Load model and project code from :mod:`message_data`."""

import re
from importlib import util
from importlib.abc import MetaPathFinder
from importlib.machinery import ModuleSpec, SourceFileLoader


class MessageDataFinder(MetaPathFinder):
    """Load model and project code from :mod:`message_data`."""

    # Expression for supported module names
    expr = re.compile(r"message_ix_models\.(?P<name>(model|project)\..*)")

    # NB coverage is excluded, because the message-ix-models test suite does not
    #    install/use message-data.
    @classmethod
    def find_spec(cls, name, path, target=None):  # pragma: no cover
        match = cls.expr.match(name)
        try:
            # Construct the name for the actual module to load
            new_name = f"message_data.{match.group('name')}"
        except AttributeError:
            # `match` was None; unsupported. Let the importlib defaults take over.
            return None

        # Get an import spec for the message_data submodule
        spec = util.find_spec(new_name)
        if not spec:
            return None

        # Create a new spec that loads message_data.model.foo as if it were
        # message_ix_models.model.foo
        new_spec = ModuleSpec(
            name=name,
            # Create a new loader that loads from the actual file with the desired name
            loader=SourceFileLoader(fullname=name, path=spec.origin),
            origin=spec.origin,
        )
        # These can't be passed through the constructor
        new_spec.submodule_search_locations = spec.submodule_search_locations

        return new_spec



import logging
import re

import pandas as pd
import pytest
from message_ix import make_df
from message_ix.testing import make_dantzig
from pandas.testing import assert_frame_equal
from sdmx.model.v21 import Code

from message_ix_models import ScenarioInfo, Spec
from message_ix_models.model.structure import get_codes, process_technology_codes
from message_ix_models.util import as_codes


class TestScenarioInfo:
    def test_empty(self):
        """ScenarioInfo created from scratch."""
        info = ScenarioInfo()

        # Set values directly
        info.set["node"] = [Code(id="AT", name="Austria")]
        info.set["year"] = [1000, 1010, 1020, 1030]
        info.y0 = 1010

        # Shorthand properties

        # `yv_ya` is generated
        assert_frame_equal(
            pd.DataFrame(
                [
                    [1010, 1010],
                    [1010, 1020],
                    [1010, 1030],
                    [1020, 1020],
                    [1020, 1030],
                    [1030, 1030],
                ],
                columns=["year_vtg", "year_act"],
            ),
            info.yv_ya,
        )

        # List of Codes is converted to list of strings
        assert ["AT"] == info.N

        # Only years >= y0
        assert [1010, 1020, 1030] == info.Y

    def test_units(self, caplog):
        """Test both :meth:`.io_units` and :meth:`.units_for`."""
        # Prepare ScenarioInfo with some commodities and technologies
        info = ScenarioInfo()
        info.set["commodity"] = get_codes("commodity")
        # NB create a technology with units annotation, since technology.yaml lacks
        #    units as of 2022-07-20
        t = as_codes({"example tech": {"units": "coulomb"}, "example tech 2": {}})
        process_technology_codes(t)
        info.set["technology"].extend(t)

        # units_for() runs, produces energy units
        c_units = info.units_for("commodity", "electr")
        assert {"[length]": 2, "[mass]": 1, "[time]": -2} == c_units.dimensionality

        # units_for() runs, produces expected units
        t_units = info.units_for("technology", "example tech")
        assert {"[current]": 1, "[time]": 1} == t_units.dimensionality

        # ValueError is raised for invalid input
        with pytest.raises(ValueError):
            info.units_for("commodity", "not a commodity")

        # io_units() runs, produces a ratio of commodity / technology units
        with caplog.at_level(logging.DEBUG, "message_ix_models"):
            assert (c_units / t_units) == info.io_units(
                "example tech", "electr", level="useful"
            )
        # level= keyword argument → logged warning
        assert "level = 'useful' ignored" == caplog.messages[-1]

        # io_units
        with pytest.raises(
            ValueError, match=re.escape("technology='example tech 2' [None]")
        ):
            info.io_units("example tech 2", "electr")

    def test_iter(self) -> None:
        info = ScenarioInfo(model="m", scenario="s")

        # dict() operates on the instance via __iter__
        assert dict(model="m", scenario="s", version=None) == dict(info)

        # Individual attributes are accessible
        assert "m" == info.model
        assert "s" == info.scenario
        assert None is info.version

    def test_url(self) -> None:
        info = ScenarioInfo(model="m", scenario="s", version=42)
        assert "m/s#42" == info.url

        info.url = "a/b"
        assert dict(model="a", scenario="b", version=None) == dict(info)

    def test_from_scenario(self, test_context) -> None:
        """ScenarioInfo initialized from an existing Scenario."""
        mp = test_context.get_platform()
        scenario = make_dantzig(mp, multi_year=True)

        # ScenarioInfo can be initialized from the scenario
        info = ScenarioInfo(scenario)

        # model, scenario, and version attributes are retrieved from `scenario`
        assert dict(
            model="Canning problem (MESSAGE scheme)", scenario="multi-year", version=1
        ) == dict(info)

        # Shorthand properties
        assert_frame_equal(
            pd.DataFrame(
                [
                    [1962, 1963],
                    [1962, 1964],
                    [1962, 1965],
                    [1963, 1963],
                    [1963, 1964],
                    [1963, 1965],
                    [1964, 1964],
                    [1964, 1965],
                    [1965, 1965],
                ],
                columns=["year_vtg", "year_act"],
            ),
            info.yv_ya,
        )
        assert [
            "World",
            "seattle",
            "san-diego",
            "new-york",
            "chicago",
            "topeka",
        ] == info.N
        assert 1963 == info.y0
        assert [1963, 1964, 1965] == info.Y

    def test_from_url(self):
        si = ScenarioInfo.from_url("m/s#123")
        assert "m" == si.model
        assert "s" == si.scenario
        assert 123 == si.version

    @pytest.mark.parametrize(
        "input, expected",
        (
            (
                "Mix-G 1.1-BM-R12 (NAV)/NPi-ref EN_20C_step-3+B#3",
                "Mix-G 1.1-BM-R12 (NAV)_NPi-ref EN_20C_step-3+B_v3",
            ),
            ("foo<>bar/baz|qux*#42", "foo_bar_baz_qux_v42"),
        ),
    )
    def test_path(self, input, expected) -> None:
        si = ScenarioInfo()
        si.url = input
        assert expected == si.path

    def test_repr(self):
        si = ScenarioInfo()
        si.set["foo"] = [1, 2, 3]
        assert "<ScenarioInfo: 3 code(s) in 1 set(s)>" == repr(si)

    def test_update(self):
        si = ScenarioInfo()
        si.par["demand"] = make_df("demand")

        # update() fails
        with pytest.raises(NotImplementedError):
            ScenarioInfo().update(si)

    @pytest.mark.parametrize(
        "codelist, y0, N_all, N_Y, y_m1, dp_checks",
        [
            (
                "A",
                2020,
                16,
                10,
                2110,
                ((1990, 10), (2010, 10), (2020, 10), (2050, 10), (2110, 10)),
            ),
            (
                "B",
                2020,
                28,
                14,
                2110,
                ((1990, 5), (2010, 5), (2020, 5), (2055, 5), (2110, 10)),
            ),
        ],
    )
    def test_year_from_codes(self, caplog, codelist, y0, N_all, N_Y, y_m1, dp_checks):
        caplog.set_level(logging.DEBUG, logger="message_ix_models")

        info = ScenarioInfo()
        codes = get_codes(f"year/{codelist}")
        info.year_from_codes(codes)

        # First model period
        assert y0 == info.y0
        assert ("firstmodelyear", y0) in info.set["cat_year"]

        # Total number of periods
        assert N_all == len(info.set["year"])

        # Number of model periods
        assert N_Y == len(info.Y)

        # Final period
        assert y_m1 == info.Y[-1]

        # Convert the data frame to a series
        dp = info.par["duration_period"].set_index("year")["value"]

        # duration_period entries are as expected
        for key, expected in dp_checks:
            assert expected == dp[key]

        # Test logging
        assert 0 == len(caplog.messages)

        info.year_from_codes(codes)

        assert 3 == len(caplog.messages)
        assert all(msg.startswith("Discard existing") for msg in caplog.messages)


class TestSpec:
    def test_getitem(self):
        s = Spec()
        with pytest.raises(KeyError):
            s["foo"]

    def test_setitem(self):
        s = Spec()

        # Setting a valid key via attribute access syntax works
        s.add = ScenarioInfo()
        # Setting a valid key via item access syntax works
        s["add"] = ScenarioInfo()

        # Setting an invalid key raises KeyError
        with pytest.raises(KeyError):
            s["foo"] = ScenarioInfo()

    def test_merge(self):
        s1 = Spec()
        s1.add.set["technology"] = ["t1", "t3", "t5"]
        s2 = Spec()
        s2.add.set["technology"] = ["t2", "t4", "t6"]

        s3 = Spec.merge(s1, s2)
        assert 6 == len(s3.add.set["technology"])



"""Tools for modeling workflows."""

import logging
import re
from typing import (
    TYPE_CHECKING,
    Callable,
    List,
    Literal,
    Mapping,
    Optional,
    Tuple,
    Union,
)

from genno import Computer
from message_ix import Scenario

from message_ix_models.util.context import Context
from message_ix_models.util.ixmp import parse_url

if TYPE_CHECKING:
    from click import Command

log = logging.getLogger(__name__)

# commented: this conflicts with option keyword arguments to workflow step functions
# CallbackType = Callable[[Context, Scenario], Scenario]
CallbackType = Callable


class WorkflowStep:
    """Single step in a multi-scenario workflow.

    Nothing occurs when the WorkflowStep is instantiated.

    Parameters
    ----------
    name : str
        ``"model name/scenario name"`` for the |Scenario| produced by the step.
    action : CallbackType, *optional*
        Function to be executed to modify the base into the target Scenario.
    clone : bool, *optional*
        :obj:`True` to clone the base scenario the target.
    target : str, *optional*
        URL for the scenario produced by the workflow step. Parsed to
        :attr:`scenario_info` and :attr:`platform_info`.
    kwargs
        Keyword arguments for `action`.
    """

    #: Function to be executed on the subject scenario. If :obj:`None`, the target
    #: scenario is loaded via :meth:`Context.get_scenario`.
    action: Optional[CallbackType] = None

    #: :obj:`True` or a :class:`dict` with keyword arguments to clone before
    #: :attr:`action` is executed. Default: :obj:`False`, do not clone.
    clone: Union[bool, dict] = False

    #: Keyword arguments passed to :attr:`action`.
    kwargs: dict

    #: Target platform name and additional options.
    platform_info: dict

    #: Target model name, scenario name, and optional version.
    scenario_info: dict

    def __init__(
        self, action: Optional[CallbackType], target=None, clone=False, **kwargs
    ):
        try:
            # Store platform and scenario info by parsing the `target` URL
            self.platform_info, self.scenario_info = parse_url(target)
        except (AttributeError, ValueError):
            if clone is not False:
                raise TypeError("target= must be supplied for clone=True")
            self.platform_info = dict()
            self.scenario_info = dict()

        # Store the callback and options
        self.action = action
        self.clone = clone
        self.kwargs = kwargs

    def __call__(
        self, context: Context, scenario: Optional[Scenario] = None
    ) -> Scenario:
        """Execute the workflow step."""
        if scenario is None:
            # No base scenario
            if self.action:
                raise RuntimeError(
                    f"Step with action {self.action!r} requires a base scenario"
                )
            # Use Context to retrieve the identified scenario
            context.platform_info.update(self.platform_info)
            context.scenario_info.update(self.scenario_info)
            s = context.get_scenario()
            log.info(f"Loaded ixmp://{s.platform.name}/{s.url}")
        else:
            # Modify the context to identify destination scenario; possibly nothing
            context.dest_scenario.update(self.scenario_info)
            s = scenario
            log.info(f"Step runs on ixmp://{s.platform.name}/{s.url}")

        if context.dest_scenario:
            log.info(f"  with context.dest_scenario={context.dest_scenario}")

        if self.clone is not False:
            # Clone to target model/scenario name
            log.info("Clone to {model}/{scenario}".format(**self.scenario_info))
            kw = self.scenario_info.copy()
            # If clone contains keyword arguments, e.g. shift_first_model_year, use them
            # NB user code should give clone = dict(keep_solution=True) if desired
            kw.update(
                self.clone
                if isinstance(self.clone, dict)
                else dict(keep_solution=False)
            )
            s = s.clone(**kw)

        if not self.action:
            return s

        log.info(f"Execute {self.action!r}")

        # Modify context to identify the target scenario
        context.set_scenario(s)

        try:
            # Invoke the callback
            result = self.action(context, s, **self.kwargs)
        except Exception:  # pragma: no cover
            s.platform.close_db()  # Avoid locking the scenario
            raise

        if result is None:
            log.info(f"…nothing returned, workflow will continue with {s.url}")
            result = s

        return result

    def __repr__(self):
        action = f"{self.action.__name__}()" if self.action else "load"
        dest = ""
        if self.scenario_info:
            dest = " -> {model}/{scenario}".format(**self.scenario_info)
        return f"<Step {action}{dest}>"


class Workflow(Computer):
    """Workflow for operations on multiple :class:`Scenarios <message_ix.Scenario>`.

    Parameters
    ----------
    context : Context
        Context object with settings common to the entire workflow.
    """

    def __init__(self, context: Context):
        super().__init__()
        self.add_single("context", context)

    def add_step(
        self,
        name: str,
        base: Optional[str] = None,
        action: Optional[CallbackType] = None,
        replace=False,
        **kwargs,
    ) -> str:
        """Add a :class:`WorkflowStep` to the workflow.

        Parameters
        ----------
        name : str
            Name for the new step.
        base : str or None
            Previous step that produces the a pre-requisite scenario for this step.
        action : CallbackType
            Function to be executed to modify the base into the target Scenario.
        replace : bool
            :data:`True` to replace an existing step.
        kwargs
            Keyword arguments for `action`; passed to and stored on the
            :class:`WorkflowStep` until used.

        Returns
        -------
        str
            The same as `name`.

        Raises
        ------
        genno.KeyExistsError
            if the step `name` already exists. Use `replace` to force overwriting an
            existing step.
        """
        # Create the workflow step
        step = WorkflowStep(action, **kwargs)

        if replace:
            # Remove any existing step
            self.graph.pop(name, None)

        # Add to the Computer; return the name of the added step
        return str(self.add_single(name, step, "context", base, strict=True))

    def run(self, name_or_names: Union[str, List[str]]):
        """Run all workflow steps necessary to produce `name_or_names`.

        Parameters
        ----------
        name_or_names: str or list of str
            Identifier(s) of steps to run.
        """
        return self.get(name_or_names)

    def truncate(self, name: str):
        """Truncate the workflow at the step `name`.

        The step `name` is replaced with a new :class:`WorkflowStep` that simply loads
        the target |Scenario| that would be produced by the original step.

        Raises
        ------
        KeyError
            if step `name` does not exist.
        """
        # Generate a new step that merely loads the scenario identified by `name` or its
        # base
        step = WorkflowStep(None)
        step.scenario_info.update(self.guess_target(name, "scenario")[0])
        try:
            step.platform_info.update(self.guess_target(name, "platform")[0])
        except KeyError as e:
            if e.args[0] is None:
                raise RuntimeError(
                    f"Unable to locate platform info for {step.scenario_info}"
                )
            else:  # pragma: no cover
                raise  # Something else

        # Replace the existing step
        self.add_single(name, step, "context", None)

    def guess_target(
        self, step_name: str, kind: Literal["platform", "scenario"] = "scenario"
    ) -> Tuple[Mapping, str]:
        """Traverse the graph looking for non-empty platform_info/scenario_info.

        Returns the info, and the step name containing it. Usually, this will identify
        the name of the platform, model, and/or scenario that is received and acted upon
        by `step_name`. This may not be the case if preceding workflow steps perform
        clone steps that are not recorded in the `target` parameter to
        :class:`WorkflowStep`.

        Parameters
        ----------
        step_name : str
           Initial step from which to work backwards.
        kind : str, "platform" or "scenario"
           Whether to look up :attr:`~WorkflowStep.platform_info` or
           :attr:`~WorkflowStep.scenario_info`.
        """
        task = self.graph[step_name]
        i = getattr(task[0], f"{kind}_info")
        return (i.copy(), step_name) if len(i) else self.guess_target(task[2], kind)


def make_click_command(wf_callback: str, name: str, slug: str, **kwargs) -> "Command":
    """Generate a click CLI command to run a :class:`.Workflow`.

    This command:

    - when invoked, imports the module containing the `wf_callback`, retrieve and calls
      the function. This function receives the values for any :mod:`click` parameters
      (arguments and/or options) passed in `kwargs`. The module is not imported
      until/unless the command is run.
    - …is automatically given the parameters:

      - :program:`--go`: Actually run the workflow; otherwise the workflow is only
        displayed.
      - :program:`--from`: Truncate the workflow at any step(s) whose names are a full
        match for this regular expression.

    - uses the :attr:`~.Computer.default_key` (if any) of the :class:`.Workflow`
      returned by `wf_callback`, if the user does not provide :program:`TARGET` on the
      command-line.

    Parameters
    ----------
    wf_callback : str
        Fully-resolved name (module and object name) for a function that generates the
        workflow; for instance "message_ix_models.project.foo.workflow.generate".
    name : str
        Descriptive workflow name used in the :program:`--help` text.
    slug : str
        File name fragment for writing the workflow diagram; the path
        :file:`{slug}-workflow.svg` is used.
    kwargs : *optional*
        Passed to :func:`click.command`, for instance to define additional parameters
        for the command.
    """
    import click

    help_arg = f"""Run the {name} workflow up to step TARGET.

    Unless --go is given, the workflow is only displayed.
    --from is interpreted as a regular expression.
    """

    @click.command(name="run", help=help_arg, **kwargs)
    @click.option("--go", is_flag=True, help="Actually run the workflow.")
    @click.option(
        "--from", "truncate_step", help="Truncate workflow at matching step(s)."
    )
    @click.argument("target_step", metavar="TARGET", required=False)
    @click.pass_obj
    def _func(context, go, truncate_step, target_step, **kwargs):
        from importlib import import_module

        from message_ix_models.util import show_versions

        # Import the module and retrieve the callback function
        module_name, callback_name = wf_callback.rsplit(".", maxsplit=1)
        module = import_module(module_name)
        callback = getattr(module, callback_name)

        # Generate the workflow
        wf = callback(context, **kwargs)

        # Truncate the workflow
        try:
            expr = re.compile(truncate_step.replace("\\", ""))
        except AttributeError:
            pass  # truncate_step is None
        else:
            for step in filter(expr.fullmatch, wf.keys()):
                log.info(f"Truncate workflow at {step!r}")
                wf.truncate(step)

        # Identify the target step
        if target_step:
            # Compile the string into a regular expression
            target_expr = re.compile(target_step)
            # Select 1 or more targets based on a regular expression in `target_step`
            target_steps = sorted(filter(lambda k: target_expr.fullmatch(k), wf.keys()))

            if len(target_steps):
                # Create a new target that collects the selected ones
                target_step = "cli-targets"
                wf.add(target_step, target_steps)
            else:
                raise click.ClickException(
                    f"No step(s) matched {target_expr!r} among:\n{sorted(wf.keys())}"
                )
        else:
            # Workflow default
            if not wf.default_key:
                raise click.ClickException(
                    f"No target step provided and no default for {wf}"
                )
            target_step = wf.default_key

        log.info(f"Execute workflow\n{wf.describe(target_step)}")
        log.debug(f"…with package versions:\n{show_versions()}")

        if not go:
            path = context.get_local_path(f"{slug}-workflow.svg")
            wf.visualize(str(path))
            log.info(f"Workflow diagram written to {path}")
            return

        wf.run(target_step)

    return _func


def solve(context, scenario, **kwargs):
    scenario.solve(**kwargs)
    return scenario



import logging

from message_ix_models.tools.exo_data import (
    ExoDataSource,
    iamc_like_data_for_query,
    register_source,
)
from message_ix_models.util import package_data_path, private_data_path

__all__ = [
    "SSPOriginal",
    "SSPUpdate",
]

log = logging.getLogger(__name__)


@register_source
class SSPOriginal(ExoDataSource):
    """Provider of exogenous data from the original SSP database.

    To use data from this source, call :func:`.exo_data.prepare_computer` with the
    arguments:

    - `source`: Any value from :data:`.SSP_2017` or equivalent string, for instance
      "ICONICS:SSP(2017).2". The specific SSP for which data is returned is determined
      from the value.
    - `source_kw` including:

      - "model": one of:

          - IIASA GDP
          - IIASA-WiC POP
          - NCAR
          - OECD Env-Growth
          - PIK GDP-32

      - "measure": The measures available differ according to the model; see the source
        data for details.

    Example
    -------
    >>> keys = prepare_computer(
    ...     context,
    ...     computer,
    ...     source="ICONICS:SSP(2015).3",
    ...     source_kw=dict(measure="POP", model="IIASA-WiC POP"),
    ... )
    >>> result = computer.get(keys[0])
    """

    id = "SSP"

    #: Name of file containing the data.
    filename = "SspDb_country_data_2013-06-12.csv.zip"

    #: One-to-one correspondence between "model" codes and date fragments in scenario
    #: codes.
    model_date = {
        "IIASA GDP": "130219",
        "IIASA-WiC POP": "130115",
        "NCAR": "130115",
        "OECD Env-Growth": "130325",
        "PIK GDP-32": "130424",
    }

    #: Replacements to apply when loading the data.
    replace = {"billion US$2005/yr": "billion USD_2005/yr"}

    def __init__(self, source, source_kw):
        s = "ICONICS:SSP(2017)."
        if not source.startswith(s):
            raise ValueError(source)

        *parts, ssp_id = source.partition(s)

        # Map the `measure` keyword to a string appearing in the data
        measure = {
            "GDP": "GDP|PPP",
            "POP": "Population",
        }[source_kw.pop("measure")]

        # Store the model ID, if any
        model = source_kw.pop("model", None)

        # Determine the date based on the model ID. There is a 1:1 correspondence.
        date = self.model_date[model]

        self.raise_on_extra_kw(source_kw)

        # Assemble a query string
        extra = "d" if ssp_id == "4" and model == "IIASA-WiC POP" else ""
        self.query = (
            f"SCENARIO == 'SSP{ssp_id}{extra}_v9_{date}' and VARIABLE == '{measure}'"
            + (f" and MODEL == '{model}'" if model else "")
        )
        # log.debug(query)

        # Iterate over possible locations for the data file
        dirs = [private_data_path("ssp"), package_data_path("test", "ssp")]
        for path in [d.joinpath(self.filename) for d in dirs]:
            if not path.exists():
                log.info(f"Not found: {path}")
                continue
            if "test" in path.parts:
                log.warning(f"Reading random data from {path}")
            break

        self.path = path

    def __call__(self):
        # Use prepared path, query, and replacements
        return iamc_like_data_for_query(self.path, self.query, replace=self.replace)


@register_source
class SSPUpdate(ExoDataSource):
    """Provider of exogenous data from the SSP Update database.

    To use data from this source, call :func:`.exo_data.prepare_computer` with the
    arguments:

    - `source`: Any value from :data:`.SSP_2024` or equivalent string, for instance
      "ICONICS:SSP(2024).2".
    - `release`: One of "3.0" or "preview".

    Example
    -------
    >>> keys = prepare_computer(
    ...     context,
    ...     computer,
    ...     source="ICONICS:SSP(2024).3",
    ...     source_kw=dict(measure="GDP", model="IIASA GDP 2023"),
    ... )
    >>> result = computer.get(keys[0])
    """

    id = "SSP update"

    #: File names containing the data, according to the release.
    filename = {
        "3.0": "1706548837040-ssp_basic_drivers_release_3.0_full.csv.gz",
        "preview": "SSP-Review-Phase-1.csv.gz",
    }

    def __init__(self, source, source_kw):
        s = "ICONICS:SSP(2024)."
        if not source.startswith(s):
            raise ValueError(source)

        *parts, ssp_id = source.partition(s)

        # Map the `measure` keyword to a 'Variable' dimension code
        measure = {
            "GDP": "GDP|PPP",
            "POP": "Population",
        }[source_kw.pop("measure")]

        # Store the model code, if any
        model = source_kw.pop("model", None)

        # Identify the data release date/version/label
        release = source_kw.pop("release", "3.0")

        self.raise_on_extra_kw(source_kw)

        # Replacements to apply, if any
        self.replace = {}

        # Prepare query pieces
        models = []
        scenarios = []

        if release == "3.0":
            # Directories in which to locate `self.filename`; stored directly within
            # message_ix_models
            dirs = [package_data_path("ssp")]

            scenarios.append(f"SSP{ssp_id}")

            if measure == "GDP|PPP":
                # Configure to prepend (m="OECD…", s="Historical Reference")
                # observations to series
                models.extend({model, "OECD ENV-Growth 2023"})
                scenarios.append("Historical Reference")
                self.replace.update(
                    Model={"OECD ENV-Growth 2023": model},
                    Scenario={"Historical Reference": scenarios[0]},
                )
        elif release == "preview":
            # Look first in message_data, then in message_ix_models test data
            dirs = [private_data_path("ssp"), package_data_path("test", "ssp")]

            scenarios.append(f"SSP{ssp_id} - Review Phase 1")
        else:
            log.error(
                f"{release = } invalid for {type(self)}; expected one of: "
                f"{set(self.filename)}"
            )
            raise ValueError(release)

        # Assemble and store a query string
        self.query = f"Scenario in {scenarios!r} and Variable == '{measure}'" + (
            f"and Model in {models!r}" if models else ""
        )
        # log.info(f"{self.query = }")

        # Iterate over possible locations for the data file
        for path in [d.joinpath(self.filename[release]) for d in dirs]:
            if not path.exists():
                log.info(f"Not found: {path}")
                continue
            if "test" in path.parts:
                log.warning(f"Reading random data from {path}")
            break

        self.path = path

    def __call__(self):
        # Use prepared path, query, and replacements
        return iamc_like_data_for_query(self.path, self.query, replace=self.replace)



"""Handle data from the ADVANCE project."""

import logging
from pathlib import Path
from typing import Optional
from zipfile import ZipFile

import pandas as pd
import pint
from genno import Quantity

from message_ix_models.project.advance.data import LOCATION, NAME
from message_ix_models.util import (
    cached,
    local_data_path,
    maybe_query,
    private_data_path,
)

log = logging.getLogger(__name__)

#: Standard dimensions for data produced as snapshots from the IIASA ENE Program
#: “WorkDB”.
DIMS = ["model", "scenario", "region", "variable", "unit", "year"]


@cached
def get_advance_data(query: Optional[str] = None) -> pd.Series:
    """Return data from the ADVANCE Work Package 2 data snapshot at :data:`LOCATION`.

    .. deprecated:: 2023.11
       Use :class:`.ADVANCE` through :func:`.exo_data.prepare_computer` instead.

    Parameters
    ----------
    query : str, optional
        Passed to :meth:`pandas.DataFrame.query` to limit the returned values.

    Returns
    -------
    pandas.Series
        with a :class:`pandas.MultiIndex` having the levels :data:`.DIMS`.
    """
    try:
        path = private_data_path(*LOCATION)
    except TypeError:
        path = local_data_path(*LOCATION)

    return _read_workdb_snapshot(path, NAME).pipe(maybe_query, query)


def advance_data(variable: str, query: Optional[str] = None) -> Quantity:
    """Return a single ADVANCE data `variable` as a :class:`genno.Quantity`.

    .. deprecated:: 2023.11
       Use :class:`.ADVANCE` through :func:`.exo_data.prepare_computer` instead.

    Parameters
    ----------
    query : str, optional
        Passed to :func:`get_advance_data`.

    Returns
    -------
    genno.Quantity
        with the dimensions :data:`.DIMS` and name `variable`. If the units of the data
        for `variable` are consistent and parseable by :mod:`pint`, the returned
        Quantity has these units; otherwise units are discarded and the returned
        Quantity is dimensionless.
    """
    data = (
        get_advance_data(query)
        .rename("value")
        .xs(variable, level="variable")
        .reset_index("unit")
    )
    if len(data.unit.unique()) > 1:  # pragma: no cover
        log.info(f"Non-unique units for {variable!r}; discarded")
        units = ""
    else:
        units = data.unit.iloc[0]

    result = Quantity(data["value"], name=variable)

    try:
        result.units = units
    except pint.errors.PintError as e:  # pragma: no cover
        log.info(f'"{e}" when parsing {units!r}; discarded')

    return result


@cached
def _read_workdb_snapshot(path: Path, name: str) -> pd.Series:
    """Read the data file.

    The expected format is a ZIP archive at `path` containing a member at `name` in CSV
    format, with columns corresponding to :data:`DIMS`, except for “year”, which is
    stored as column headers (‘wide’ format). (This corresponds to an older version of
    the “IAMC format,” without more recent additions intended to represent sub-annual
    time resolution using a separate column.)

    .. deprecated:: 2023.11
       Use :func:`.iamc_like_data_for_query` instead.
    """
    with ZipFile(path) as zf:  # Open the ZIP archive
        with zf.open(name) as f:  # Open a particular member
            # - Read data using upper case column names, then convert to lower-case.
            # - Drop null rows.
            # - Stack the “year” dimension (‘long’ format), creating a pd.Series.
            # - Apply the index names.
            return (
                pd.read_csv(f, index_col=list(map(str.upper, DIMS[:-1])))
                .rename(columns=lambda c: int(c))
                .dropna(how="all")
                .stack()
                .rename_axis(DIMS)
            )



"""Command-line utilities.

These are used for building CLIs using :mod:`click`.
"""

import logging
import sys
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Callable, List, Literal, Mapping, Optional, Union, cast

import click
import click.testing
from click import Argument, Choice, Option

from message_ix_models import Context, model
from message_ix_models.model.structure import codelists

from ._logging import preserve_log_level
from .scenarioinfo import ScenarioInfo

log = logging.getLogger(__name__)


def common_params(param_names: str):
    """Decorate a click.command with common parameters `param_names`.

    `param_names` must be a space-separated string of names appearing in :data:`PARAMS`,
    e.g. ``"ssp force output_model"``. The decorated function receives keyword
    arguments with these names::

        @click.command()
        @common_params("ssp force output_model")
        def mycmd(ssp, force, output_model)
            # ...
    """

    # Simplified from click.decorators._param_memo
    def decorator(f):
        if not hasattr(f, "__click_params__"):
            f.__click_params__ = []
        f.__click_params__.extend(
            PARAMS[name] for name in reversed(param_names.split())
        )
        return f

    return decorator


def default_path_cb(*default_parts):
    """Return a callback function for click.Option handling.

    If no option value is given, the callback uses :meth:`Context.get_local_path` and
    `default_parts` to provide a path that is relative to local data directory, e.g.
    the current working directory (see :doc:`/data`).
    """

    def _callback(context, param, value):
        value = value or context.obj.get_local_path(*default_parts)
        setattr(context.obj, param.name, value)
        return value

    return _callback


def exec_cb(expression: str) -> Callable:
    """Return a callback that :func:`exec`-utes an `expression`.

    The `expression` is executed in a limited context that has only two names available:

    - :py:`context`: the :class:`.Context` instance.
    - :py:`value`: the value passed to the :mod:`click.Parameter`.

    Example
    -------
    >>> @click.command
    ... @click.option(
    ...     "--myopt", callback=exec_cb("context.my_mod.my_opt = value + 3")
    ... )
    ... def cmd(...):
    ...     ...
    """

    def _cb(context: Union[click.Context, Context], param, value):
        ctx = context.obj if isinstance(context, click.Context) else context
        exec(expression, {}, {"context": ctx, "value": value})
        return value

    return _cb


def format_sys_argv() -> str:
    """Format :data:`sys.argv` in a readable manner."""
    lines = ["Invoked:"]
    indent = ""
    for item in sys.argv:
        lines.append(f"{indent}{item} \\")
        indent = "  "

    return "\n".join(lines)[:-2]


def store_context(context: Union[click.Context, Context], param, value):
    """Callback that simply stores a value on the :class:`.Context` object.

    Use this for parameters that are not used directly in a @click.command() function,
    but need to be carried by the Context for later use.
    """
    setattr(
        context.obj if isinstance(context, click.Context) else context,
        param.name,
        value,
    )
    return value


@contextmanager
def temporary_command(group: "click.Group", command: "click.Command"):
    """Temporarily attach command `command` to `group`."""
    assert command.name is not None
    try:
        group.add_command(command)
        yield
    finally:
        group.commands.pop(command.name)


def urls_from_file(
    context: Union[click.Context, Context], param, value
) -> List[ScenarioInfo]:
    """Callback to parse scenario URLs from `value`."""
    si: List[ScenarioInfo] = []

    if value is None:
        return si

    with click.open_file(value) as f:
        for line in f:
            si.append(ScenarioInfo.from_url(url=line))

    # Store on context
    mm_context = context.obj if isinstance(context, click.Context) else context
    mm_context.core.scenarios = si

    return si


def unique_id() -> str:
    """Return a unique ID for a CLI invocation.

    The return value resembles "mix-models-debug-3332d415ef65bf2a-2023-02-02T162931" and
    contains:

    - The CLI name and (sub)commands.
    - A hash of all the CLI parameters (options and arguments).
    - The current date and time, in ISO format with Windows-incompatible ":" removed.

    """
    click_context = click.get_current_context()

    # Collapse CLI (sub)commands and their arguments to a hashable data structure
    # This also includes CLI options *not* given
    c: Optional[click.Context] = click_context
    data = []
    while c is not None:
        data.append((c.command.name, tuple(sorted(c.params.items()))))
        c = c.parent

    # Assemble parts
    return "-".join(
        [
            click_context.command_path.replace(" ", "-"),
            f"{hash(tuple(reversed(data))):x}",
            datetime.now().isoformat(timespec="seconds").replace(":", ""),
        ]
    )


#: Common command-line parameters (arguments and options). See :func:`common_params`.
PARAMS = {
    "dest": Option(
        ["--dest"],
        callback=store_context,
        expose_value=False,
        help="Destination URL for created scenario(s).",
    ),
    "dry_run": Option(
        ["--dry-run"],
        is_flag=True,
        callback=store_context,
        expose_value=False,
        help="Only show what would be done.",
    ),
    "force": Option(
        ["--force"],
        is_flag=True,
        callback=store_context,
        help="Overwrite or modify existing model/scenario.",
    ),
    "nodes": Option(
        ["--nodes"],
        help="Code list to use for 'node' dimension.",
        callback=exec_cb("context.model.regions = value"),
        type=Choice(codelists("node")),
        default=model.Config.regions,
        expose_value=False,
    ),
    "output_model": Option(
        ["--output-model"], help="Model name under which scenarios should be generated."
    ),
    "policy_path": Option(
        ["--policy-path"],
        callback=default_path_cb("scenario_generation", "policies"),
        help="Path to policy scripts.",
    ),
    "platform_dest": Option(["--platform-dest"], help="Name of destination Platform."),
    "quiet": Option(
        ["--quiet"],
        is_flag=True,
        expose_value=False,
        help="Show less or no output.",
    ),
    "regions": Option(
        ["--regions"],
        help="Code list to use for 'node' dimension.",
        callback=exec_cb("context.model.regions = value or context.model.regions"),
        type=Choice(codelists("node")),
    ),
    "rep_out_path": Option(
        ["--rep-out-path"],
        callback=default_path_cb("reporting_output"),
        help="Path for reporting output.",
    ),
    "rep_template": Option(
        ["--rep-template"],
        callback=default_path_cb(
            "message_data", "tools", "post_processing", "MESSAGEix_WorkDB_Template.xlsx"
        ),
        help="Path incl. filename and extension to reporting template.",
    ),
    "run_reporting_only": Option(
        ["--run-reporting-only"],
        is_flag=True,
        callback=store_context,
        help="Run only reporting.",
    ),
    "ssp": Argument(
        ["ssp"], callback=store_context, type=Choice(["SSP1", "SSP2", "SSP3"])
    ),
    "urls_from_file": Option(
        ["--urls-from-file", "-f"],
        type=click.Path(
            exists=True,
            dir_okay=False,
            resolve_path=True,
            allow_dash=True,
            path_type=Path,
        ),
        callback=urls_from_file,
    ),
    "verbose": Option(
        # NB cannot use store_callback here; this is processed in the top-level CLI
        #    before the message_ix_models.Context() object is set up
        ["--verbose", "-v"],
        is_flag=True,
        help="Print DEBUG-level log messages.",
    ),
    "years": Option(
        ["--years"],
        help="Code list to use for the 'year' dimension.",
        callback=exec_cb("context.model.years = value"),
        type=Choice(codelists("year")),
        default=model.Config.years,
        # expose_value=False,
    ),
}


@dataclass
class CliRunner:
    """Similar to :class:`click.testing.CliRunner`, with extra features."""

    #: CLI entry point
    cli_cmd: click.Command
    #: CLI module
    cli_module: str

    env: Mapping[str, str] = field(default_factory=dict)
    charset: str = "utf-8"

    #: Method for invoking the command
    method: Literal["click", "subprocess"] = "click"

    def invoke(self, *args, **kwargs) -> click.testing.Result:
        method = kwargs.pop("method", self.method)

        if method == "click":
            runner = click.testing.CliRunner(env=self.env)
            with preserve_log_level():
                result = runner.invoke(self.cli_cmd, *args, **kwargs)
        elif method == "subprocess":
            result = self.invoke_subprocess(*args, **kwargs)

        # Store the result to be used by assert_exit_0()
        self.last_result = result

        return result

    def invoke_subprocess(self, *args, **kwargs) -> click.testing.Result:
        """Invoke the CLI in a subprocess."""
        import subprocess

        assert 1 == len(args)
        all_args: List[str] = [sys.executable, "-m", self.cli_module, *args[0]]

        # Run; capture in a subprocess.CompletedProcess
        cp = subprocess.run(all_args, capture_output=True, env=self.env, **kwargs)

        # Convert to a click.testing.Result
        return click.testing.Result(
            runner=cast(click.testing.CliRunner, self),
            stdout_bytes=cp.stdout or bytes(),
            stderr_bytes=cp.stderr or bytes(),
            return_value=None,
            exit_code=cp.returncode,
            exception=None,
            exc_info=None,
        )

    def assert_exit_0(self, *args, **kwargs) -> click.testing.Result:
        """Assert a result has exit_code 0, or print its traceback.

        If any `args` or `kwargs` are given, :meth:`.invoke` is first called. Otherwise,
        the result from the last call of :meth:`.invoke` is used.

        Raises
        ------
        AssertionError
            if the result exit code is not 0.
        """
        __tracebackhide__ = True

        if len(args) + len(kwargs):
            self.invoke(*args, **kwargs)

        # Retrieve the last result
        result = self.last_result

        if result.exit_code != 0:
            print(f"{result.exit_code = }", f"{result.output = }", sep="\n")
            raise RuntimeError(result.exit_code)

        return result



"""Basic tests of the command line."""

import ixmp
import pytest
from message_ix.testing import make_dantzig

from message_ix_models import util

COMMANDS = [
    tuple(),
    ("debug",),
    ("fetch",),
    ("report",),
    ("ssp", "gen-structures"),
    ("techs",),
    ("testing",),
    ("water-ix",),
]


def _cli_help_id(argvalue):
    return f"mix-models {' '.join(argvalue)} --help"


@pytest.mark.parametrize("command", COMMANDS, ids=_cli_help_id)
def test_cli_help(mix_models_cli, command):
    """--help works for every CLI command."""
    mix_models_cli.assert_exit_0(list(command) + ["--help"])


def test_cli_debug(mix_models_cli):
    """The 'debug' CLI command can be invoked."""
    mix_models_cli.assert_exit_0(["debug"])


def test_cli_export_test_data(monkeypatch, tmp_path, test_context, mix_models_cli):
    """The :command:`export-test-data` command can be invoked."""
    # Create an empty scenario in the temporary local file database
    platform = "local"
    mp = ixmp.Platform(platform)
    scen = make_dantzig(mp)

    # URL
    url = f"ixmp://{platform}/{scen.model}/{scen.scenario}#{scen.version}"

    # Monkeypatch MESSAGE_DATA_PATH in case tests are being performed on a system
    # without message_data installed
    monkeypatch.setattr(util.common, "MESSAGE_DATA_PATH", tmp_path)
    tmp_path.joinpath("data", "tests").mkdir(exist_ok=True, parents=True)

    # File that will be created
    technology = ["coal_ppl"]
    dest_file = util.private_data_path(
        "tests", f"{scen.model}_{scen.scenario}_{'_'.join(technology)}.xlsx"
    )

    # Release the database lock
    mp.close_db()

    # Export works
    result = mix_models_cli.assert_exit_0([f"--url={url}", "export-test-data"])

    # The file is created in the expected location
    assert str(dest_file) in result.output
    assert dest_file.exists()



import pandas as pd
import pandas.testing as pdt
import pytest
from sdmx.model import Annotation, Code

from message_ix_models.model.macro import generate, load
from message_ix_models.util import package_data_path


@pytest.mark.parametrize(
    "parameter, value",
    [
        ("aeei", 1.0),
        ("config", None),
        ("depr", 1.0),
        ("drate", 1.0),
        ("lotol", 1.0),
        pytest.param("foo", 1.0, marks=pytest.mark.xfail(raises=NotImplementedError)),
    ],
)
def test_generate0(test_context, parameter, value):
    result = generate(parameter, test_context, value=value)

    assert not result.isna().any(axis=None)


def test_generate1(test_context):
    commodities = [
        Code(id="foo", annotations=[Annotation(id="macro-sector", text="BAR")]),
        Code(id="baz", annotations=[Annotation(id="macro-sector", text="QUX")]),
    ]

    result = generate("config", test_context, commodities)

    assert {"foo", "baz"} == set(result["commodity"].unique())

    # Only the identified sectors appear
    assert {"BAR", "QUX"} == set(result["sector"].unique())

    # Only 2 unique (commodity, sector) combinations appear
    assert 2 == len(result[["commodity", "sector"]].drop_duplicates())


def test_load(test_context):
    result = load(package_data_path("test", "macro"))
    assert {"kgdp"} == set(result.keys())
    pdt.assert_index_equal(pd.Index(["node", "value", "unit"]), result["kgdp"].columns)
    assert not result["kgdp"].isna().any(axis=None)



"""Generate input data."""

import logging

from message_ix_models import ScenarioInfo
from message_ix_models.util import add_par_data

from .demands import add_irrigation_demand, add_sectoral_demands, add_water_availability
from .infrastructure import add_desalination, add_infrastructure_techs
from .irrigation import add_irr_structure
from .water_for_ppl import cool_tech, non_cooling_tec
from .water_supply import add_e_flow, add_water_supply

log = logging.getLogger(__name__)

DATA_FUNCTIONS = [
    add_water_supply,
    cool_tech,  # Water & parasitic_electricity requirements for cooling technologies
    non_cooling_tec,
    add_sectoral_demands,
    add_water_availability,
    add_irrigation_demand,
    add_infrastructure_techs,
    add_desalination,
    add_e_flow,
    add_irr_structure,
]

DATA_FUNCTIONS_COUNTRY = [
    add_water_supply,
    cool_tech,  # Water & parasitic_electricity requirements for cooling technologies
    non_cooling_tec,
    add_sectoral_demands,
    add_water_availability,
    # add_irrigation_demand, # not used and coming from GLOBIOM for the global region
    add_infrastructure_techs,
    add_desalination,
    add_e_flow,
    # add if statement: if irrigation: land component from external model
]


def add_data(scenario, context, dry_run=False):
    """Populate `scenario` with MESSAGEix-Nexus data."""

    info = ScenarioInfo(scenario)
    context["water build info"] = info

    data_funcs = (
        [add_water_supply, cool_tech, non_cooling_tec]
        if context.nexus_set == "cooling"
        else DATA_FUNCTIONS
        if context.type_reg == "global"
        else DATA_FUNCTIONS_COUNTRY
    )

    for func in data_funcs:
        # Generate or load the data; add to the Scenario
        log.info(f"from {func.__name__}()")
        add_par_data(scenario, func(context), dry_run=dry_run)

    log.info("done")



import re

import ixmp
import message_ix
import pandas as pd
import pandas.testing as pdt
import pytest
import xarray as xr
from genno import Computer, Quantity
from ixmp.testing import assert_logs
from message_ix.testing import make_dantzig

from message_ix_models import ScenarioInfo
from message_ix_models.model.structure import get_codes
from message_ix_models.report.operator import (
    compound_growth,
    filter_ts,
    from_url,
    get_ts,
    gwp_factors,
    make_output_path,
    model_periods,
    remove_ts,
    share_curtailment,
)

from ..test_report import MARK


@pytest.fixture
def c() -> Computer:
    return Computer()


@pytest.fixture
def scenario(test_context):
    mp = test_context.get_platform()
    yield make_dantzig(mp)


def test_compound_growth():
    """:func:`.compound_growth` on a 2-D quantity."""
    qty = Quantity(
        xr.DataArray(
            [
                [1.01, 1.0, 1.02, 1e6],  # Varying growth rates for x=x1
                [1.0, 1.0, 1.0, 1.0],  # No rates/constant for x=x2
            ],
            coords=(["x1", "x2"], [2020, 2021, 2030, 2035]),
            dims=("x", "t"),
        )
    )

    # Function runs
    result = compound_growth(qty, "t")

    # Results have expected values
    r1 = result.sel(x="x1")
    assert all(1.0 == r1.sel(t=2020))
    assert all(1.01 == r1.sel(t=2021) / r1.sel(t=2020))
    assert all(1.0 == r1.sel(t=2030) / r1.sel(t=2021))
    assert all(1.02**5 == r1.sel(t=2035) / r1.sel(t=2030))

    assert all(1.0 == result.sel(x="x2"))


def test_filter_ts():
    df = pd.DataFrame([["foo"], ["bar"]], columns=["variable"])
    assert 2 == len(df)

    # Operator runs
    result = filter_ts(df, re.compile(".(ar)"))

    # Only matching rows are returned
    assert 1 == len(result)

    # Only the first match group in `expr` is preserved
    assert {"ar"} == set(result.variable.unique())


@MARK[0]
def test_from_url(scenario):
    full_url = f"ixmp://{scenario.platform.name}/{scenario.url}"

    # Operator runs
    result = from_url(full_url)
    # Result is of the default class
    assert result.__class__ is ixmp.TimeSeries
    # Same object was retrieved
    assert scenario.url == result.url

    # Same, but specifying message_ix.Scenario
    result = from_url(full_url, message_ix.Scenario)
    assert result.__class__ is message_ix.Scenario
    assert scenario.url == result.url


@MARK[0]
def test_get_remove_ts(caplog, scenario):
    # get_ts() runs
    result0 = get_ts(scenario)
    pdt.assert_frame_equal(scenario.timeseries(), result0)

    # Can be used through a Computer

    c = Computer()
    c.require_compat("message_ix_models.report.operator")
    c.add("scenario", scenario)

    key = c.add("test1", "get_ts", "scenario", filters=dict(variable="GDP"))
    result1 = c.get(key)
    assert 3 == len(result1)

    # remove_ts() can be used through Computer
    key = c.add("test2", "remove_ts", "scenario", "config", after=1964)

    # Task runs, logs
    # NB this log message is incorrect, because ixmp's JDBCBackend is unable to delete
    #    data stored with "meta=True". Only 1 row is removed
    with assert_logs(caplog, "Remove 2 of 6 (1964 <= year) rows of time series data"):
        c.get(key)

    # See comment above; only one row is removed
    assert 6 - 1 == len(scenario.timeseries())

    # remove_ts() can be used directly
    remove_ts(scenario)

    # All non-'meta' data were removed
    assert 3 == len(scenario.timeseries())


def test_gwp_factors():
    result = gwp_factors()

    assert ("gwp metric", "e", "e equivalent") == result.dims


def test_make_output_path(tmp_path, c):
    # Configure a Computer, ensuring the output_dir configuration attribute is set
    c.configure(output_dir=tmp_path)

    # Add a computation that invokes make_output_path
    c.add("test", make_output_path, "config", "foo.csv")

    # Returns the correct path
    assert tmp_path.joinpath("foo.csv") == c.get("test")


def test_model_periods():
    # Prepare input data
    si = ScenarioInfo()
    si.year_from_codes(get_codes("year/B"))
    cat_year = pd.DataFrame(si.set["cat_year"], columns=["type_year", "year"])

    # Operator runs
    result = model_periods(si.set["year"], cat_year)

    assert isinstance(result, list)
    assert all(isinstance(y, int) for y in result)
    assert 2020 == min(result)


@pytest.mark.xfail(reason="Incomplete")
def test_share_curtailment():
    share_curtailment()



"""Simulated solution data for testing :mod:`~message_ix_models.report`."""

import logging
from collections import ChainMap, defaultdict
from collections.abc import Mapping
from copy import deepcopy
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence, Union

import pandas as pd
from dask.core import quote
from genno import Key, KeyExistsError, Quantity
from message_ix import Reporter
from pandas.api.types import is_scalar

from message_ix_models import ScenarioInfo
from message_ix_models.util import minimum_version
from message_ix_models.util._logging import mark_time, silence_log
from message_ix_models.util.ixmp import rename_dims

if TYPE_CHECKING:
    from message_ix.models import Item

__all__ = [
    "add_simulated_solution",
    "data_from_file",
    "simulate_qty",
    "to_simulate",
]

log = logging.getLogger(__name__)


def dims_of(info: "Item") -> Dict[str, str]:
    """Return a mapping from the full index names to short dimension IDs of `info`."""
    return {d: rename_dims().get(d, d) for d in (info.dims or info.coords or [])}


@minimum_version("message_ix 3.7.0.post0")
@lru_cache(1)
def to_simulate():
    """Return items to be included in a simulated solution."""
    from message_ix.models import MACRO, MESSAGE

    # Items to included in a simulated solution: MESSAGE sets and parameters; some
    # variables
    result = deepcopy(MESSAGE.items)
    # MACRO variables
    result.update({k: MACRO.items[k] for k in ("GDP", "MERtoPPP")})

    return result


def simulate_qty(
    name: str, dims: List[str], item_data: Union[dict, pd.DataFrame]
) -> Quantity:
    """Return simulated data for item `name`.

    Parameters
    ----------
    dims :
        Dimensions of the resulting quantity.
    item_data :
        Optional data for the quantity.
    """
    if isinstance(item_data, dict):
        # NB this is code lightly modified from make_df

        # Default values for every column
        data: Mapping = ChainMap(item_data, defaultdict(lambda: None))

        # Arguments for pd.DataFrame constructor
        args: Dict[str, Any] = dict(data={})

        # Flag if all values in `data` are scalars
        all_scalar = True

        for column in dims + ["value"]:
            # Update flag
            all_scalar &= is_scalar(data[column])
            # Store data
            args["data"][column] = data[column]

        if all_scalar:
            # All values are scalars, so the constructor requires an index to be passed
            # explicitly.
            args["index"] = [0]

        df = pd.DataFrame(**args)
    else:
        # Provided complete data frame
        df = item_data.rename(columns=rename_dims())

    # Data must be entirely empty, or complete
    assert not df.isna().any().any() or df.isna().all().all(), data
    assert not df.duplicated().any(), f"Duplicate data for simulated {repr(name)}"

    return Quantity(df.set_index(dims)["value"] if len(dims) else df, name=name)


def data_from_file(path: Path, *, name: str, dims: Sequence[str]) -> Quantity:
    """Read simulated solution data for item `name` from `path`.

    For variables and equations (`name` in upper case), the file **must** have columns
    corresponding to `dims` followed by "Val", "Marginal", "Upper", and "Scale". The
    "Val" column is returned.

    For parameters, the file **must** have columns corresponding to `dims` followed by
    "value" and "unit". The "value" column is returned.
    """
    if name.isupper():
        # Construct a list of the columns
        # NB Must assign the dimensions directly; they cannot be read from the file, as
        #    the column headers are the internal GAMS set names (e.g. "year_all")
        #    instead of the index names from message_ix.
        cols = list(dims) + ["Val", "Marginal", "Lower", "Upper", "Scale"]

        return Quantity(
            pd.read_csv(path, engine="pyarrow")
            .set_axis(cols, axis=1)
            .set_index(cols[:-5])["Val"],
            name=name,
        )
    else:
        cols = list(dims) + ["value", "unit"]
        tmp = (
            pd.read_csv(path, engine="pyarrow")
            # Drop a leading index column that appears in some files
            # TODO Adjust .snapshot.unpack() to avoid generating this column; update
            # data; then remove this call
            .drop(columns="", errors="ignore")
            .set_axis(cols, axis=1)
            .set_index(cols[:-2])
        )
        # TODO pass units if they are unique
        return Quantity(tmp["value"], name=name)


@minimum_version("message_ix 3.6")
def add_simulated_solution(
    rep: Reporter,
    info: ScenarioInfo,
    data: Optional[Dict] = None,
    path: Optional[Path] = None,
):
    """Add a simulated model solution to `rep`.

    Parameters
    ----------
    data : dict or pandas.DataFrame, *optional*
        If given, a mapping from MESSAGE item (set, parameter, or variable) names to
        inputs that are passed to :func:`simulate_qty`.
    path : Path, *optional*
        If given, a path to a directory containing one or more files with names like
        :file:`ACT.csv.gz`. These files are taken as containing "simulated" model
        solution data for the MESSAGE variable with the same name. See
        :func:`data_from_file`.
    """
    from ixmp.backend import ItemType

    rep.configure(
        rename_dims=dict(
            node_rel="nr",
            year_rel="yr",
        ),
    )

    mark_time()
    N = len(rep.graph)

    # Ensure "scenario" is present in the graph
    rep.graph.setdefault("scenario", None)

    # Add simulated data
    data = data or dict()
    for name, item_info in to_simulate().items():
        dims = list(dims_of(item_info).values())
        key = Key(name, dims)

        # Add a task to load data from a file in `path`, if it exists
        try:
            assert path is not None
            p = path.joinpath(name).with_suffix(".csv.gz")
            assert p.exists()
        except AssertionError:
            pass  # No `path` or no such file
        else:
            # Add data from file
            rep.add(key, data_from_file, p, name=name, dims=key.dims, sums=True)
            continue

        if item_info.type == ItemType.SET:
            # Add the set elements from `info`
            rep.add(rename_dims().get(name, name), quote(info.set[name]))
        elif item_info.type in (ItemType.PAR, ItemType.VAR):
            # Retrieve an existing key for `name`
            try:
                full_key = rep.full_key(name)
            except KeyError:
                full_key = None  # Not present in `rep`

            # Simulate data for name
            item_data = data.get(name)

            if full_key and not item_data:
                # Don't overwrite existing task with empty data
                continue

            # Add a task to simulate data for this quantity
            rep.add(
                key,
                simulate_qty,
                name=name,
                dims=dims,
                item_data=item_data,
                sums=True,
            )

    log.info(f"{len(rep.graph) - N} keys")
    N = len(rep.graph)
    mark_time()

    # Prepare the base MESSAGEix computations
    with silence_log("genno", logging.CRITICAL):
        try:
            rep.add_tasks()
        except KeyExistsError:
            pass  # `rep` was produced with Reporter.from_scenario()

    log.info(f"{len(rep.graph)} total keys")
    mark_time()



"""Compatibility code that emulates :mod:`.message_data` reporting."""

import logging
from functools import partial
from itertools import chain, count
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional

from genno import Key, Quantity, quote
from genno.core.key import iter_keys, single_key

if TYPE_CHECKING:
    from genno import Computer
    from ixmp import Reporter
    from sdmx.model.common import Code

    from message_ix_models import Context

__all__ = [
    "TECH_FILTERS",
    "callback",
    "eff",
    "emi",
    "get_techs",
    "inp",
    "out",
    "prepare_techs",
]

log = logging.getLogger(__name__)

#: Filters for determining subsets of technologies.
#:
#: Each value is a Python expression :func:`eval`'d in an environment containing
#: variables derived from the annotations on :class:`Codes <.Code>` for each technology.
#: If the expression evaluates to :obj:`True`, then the code belongs to the set
#: identified by the key.
#:
#: See also
#: --------
#: get_techs
#: prepare_techs
#:
TECH_FILTERS = {
    "gas all": "c_in == 'gas' and l_in in 'secondary final' and '_ccs' not in id",
    "gas extra": "False",
    # Residential and commercial
    "trp coal": "sector == 'transport' and c_in == 'coal'",
    "trp gas": "sector == 'transport' and c_in == 'gas'",
    "trp foil": "sector == 'transport' and c_in == 'fueloil'",
    "trp loil": "sector == 'transport' and c_in == 'lightoil'",
    "trp meth": "sector == 'transport' and c_in == 'methanol'",
    # Transport
    "rc gas": "sector == 'residential/commercial' and c_in == 'gas'",
}


# Counter for anon()
_ANON = map(lambda n: Key(f"_{n}"), count())


def anon(name: Optional[str] = None, dims: Optional[Key] = None) -> Key:
    """Create an ‘anonymous’ :class:`.Key`, optionally with `dims` from another Key."""
    result = next(_ANON) if name is None else Key(name)

    return result.append(*getattr(dims, "dims", []))


def get_techs(c: "Computer", prefix: str, kinds: Optional[str] = None) -> List[str]:
    """Return a list of technologies.

    The list is assembled from lists in `c` with keys like "t::{prefix} {kind}",
    with one `kind` for each space-separated item in `kinds`. If no `kinds` are
    supplied, "t::{prefix}" is used.

    See also
    --------
    prepare_techs
    """
    _kinds = kinds.split() if kinds else [""]
    return list(chain(*[c.graph[f"t::{prefix} {k}".rstrip()][0].data for k in _kinds]))


def make_shorthand_function(
    base_name: str, to_drop: str, default_unit_key: Optional[str] = None
):
    """Create a shorthand function for adding tasks to a :class:`.Reporter`."""
    _to_drop = to_drop.split()

    def func(
        c: "Computer",
        technologies: List[str],
        *,
        name: Optional[str] = None,
        filters: Optional[dict] = None,
        unit_key: Optional[str] = default_unit_key,
    ) -> Key:
        f"""Select data from "{base_name}:*" and apply units.

        The returned key sums the result over the dimensions {_to_drop!r}.

        Parameters
        ----------
        technologies :
            List of technology IDs to include.
        name : str, *optional*
            If given, the name of the resulting key. Default: a name like "_123"
            generated with :func:`anon`.
        filters : dict, *optional*
            Additional filters for selecting data from "{base_name}:*". Keys are short
            dimension names (for instance, "c" for "commodity"); values are lists of
            IDs.
        unit_key : str, *optional*
            Key for units to apply to the result. Must appear in :attr:`.Config.units`.
        """
        base = single_key(c.full_key(base_name))
        key = anon(name, dims=base)

        indexers = dict(t=technologies)
        indexers.update(filters or {})

        if unit_key:
            c.add(key + "sel", "select", base, indexers=indexers)
            c.add(
                key,
                "assign_units",
                key + "sel",
                units=c.graph["config"]["model"].units[unit_key],
                sums=True,
            )
        else:
            c.add(key, "select", base, indexers=indexers, sums=True)

        # Return the partial sum over some dimensions
        return key.drop(*_to_drop)

    return func


inp = make_shorthand_function("in", "c h ho l no t", "energy")
emi = make_shorthand_function("rel", "nr r t yr")
out = make_shorthand_function("out", "c h hd l nd t", "energy")


def eff(
    c: "Computer",
    technologies: List[str],
    filters_in: Optional[dict] = None,
    filters_out: Optional[dict] = None,
) -> Key:
    """Throughput efficiency (input / output) for `technologies`.

    Equivalent to :meth:`PostProcess.eff`.

    Parameters
    ----------
    filters_in : dict, *optional*
        Passed as the `filters` parameter to :func:`inp`.
    filters_out : dict, *optional*
        Passed as the `filters` parameter to :func:`out`.
    """
    # TODO Check whether append / drop "t" is necessary
    num = c.graph.unsorted_key(inp(c, technologies, filters=filters_in).append("t"))
    denom = c.graph.unsorted_key(out(c, technologies, filters=filters_out).append("t"))
    assert isinstance(num, Key)
    assert isinstance(denom, Key)

    key = anon(dims=num)

    c.add(key, "div", num, denom, sums=True)

    return key.drop("t")


def pe_w_ccs_retro(
    c: "Computer",
    t: str,
    t_scrub: str,
    k_share: Optional[Key],
    filters: Optional[dict] = None,
) -> Key:
    """Calculate primary energy use of technologies with scrubbers.

    Equivalent to :func:`default_tables._pe_wCCS_retro` at L129.
    """
    ACT: Key = single_key(c.full_key("ACT"))

    k0 = out(c, [t_scrub])
    k1 = c.add(anon(), "mul", k0, k_share) if k_share else k0

    k2 = anon(dims=ACT).drop("t")
    c.add(k2, "select", ACT, indexers=dict(t=t), drop=True, sums=True)

    # TODO determine the dimensions to drop for the numerator
    k3, *_ = iter_keys(c.add(anon(dims=k2), "div", k2.drop("yv"), k2, sums=True))
    assert_dims(c, k3)

    filters_out = dict(c=["electr"], l=["secondary"])
    k4 = eff(c, [t], filters_in=filters, filters_out=filters_out)
    k5 = single_key(c.add(anon(), "mul", k3, k4))
    k6 = single_key(c.add(anon(dims=k5), "div", k1, k5))
    assert_dims(c, k6)

    return k6


def prepare_techs(c: "Computer", technologies: List["Code"]) -> None:
    """Prepare sets of technologies in `c`.

    For each `key` → `expr` in :data:`TECH_FILTERS` and each technology :class:`Code`
    `t` in `technologies`:

    - Apply the filter expression `expr` to information about `t`.
    - If the expression evaluates to :obj:`True`, add it to a list in `c` at "t::{key}".

    These lists of technologies can be used directly or retrieve with :func:`get_techs`.
    """
    result: Mapping[str, List[str]] = {k: list() for k in TECH_FILTERS}

    warned = set()  # Filters that raise some kind of Exception

    # Iterate over technologies
    for t in technologies:
        # Assemble information about `t` from its annotations
        info: Dict[str, Any] = dict(id=t.id)
        # Sector
        info["sector"] = str(t.get_annotation(id="sector").text)
        try:
            # Input commodity and level
            info["c_in"], info["l_in"] = t.eval_annotation("input")
        except (TypeError, ValueError):
            info["c_in"] = info["l_in"] = None

        # Iterate over keys and respective filters
        for key, expr in TECH_FILTERS.items():
            try:
                # Apply the filter to the `info` about `t`
                if eval(expr, None, info) is True:
                    # Filter evaluates to True → add `t` to the list of labels for `key`
                    result[key].append(t.id)
            except Exception as e:
                # Warn about this filter, only once
                if expr not in warned:
                    log.warning(f"{e!r} when evaluating {expr!r}")
                    warned.add(expr)

    # Add keys like "t::trp gas" corresponding to TECH_FILTERS["trp gas"]
    for k, v in result.items():
        c.add(f"t::{k}", quote(sorted(v)))


def assert_dims(c: "Computer", *keys: Key):
    """Check the dimensions of `keys` for an "add", "sub", or "div" task.

    This is a sanity check needed because :py:`c.add("name", "div", …)` does not (yet)
    automatically infer the dimensions of the resulting key. This is in contrast to
    :py:`c.add("name", "mul", …)`, which *does* infer.

    Use this function after manual construction of a key for a "add", "div", or "sub"
    task, in order to ensure the key matches the dimensionality of the quantity that
    will result from the task.

    .. todo:: Remove once handled upstream in :mod:`genno`.
    """
    for key in keys:
        task = c.graph[key]
        expected = Key.product("foo", *task[1:])

        op = f" {task[0].__name__} "
        assert set(key.dims) == set(expected.dims), (
            f"Task should produce {op.join(repr(k) for k in task[1:])} = "
            f"{str(expected).split(':')[1]}; key indicates {str(key).split(':')[1]}"
        )


def callback(rep: "Reporter", context: "Context") -> None:
    """Partially duplicate the behaviour of :func:`.default_tables.retr_CO2emi`.

    Currently, this prepares the following keys and the necessary preceding
    calculations:

    - "transport emissions full::iamc": data for the IAMC variable
      "Emissions|CO2|Energy|Demand|Transportation|Road Rail and Domestic Shipping"
    """
    from message_ix_models.model.bare import get_spec

    from . import iamc

    N = len(rep.graph)

    # Structure information
    spec = get_spec(context)
    prepare_techs(rep, spec.add.set["technology"])

    # Constants from report/default_units.yaml
    rep.add("conv_c2co2:", 44.0 / 12.0)  # dimensionless
    # “Carbon content of natural gas”
    rep.add("crbcnt_gas:", Quantity(0.482, units="Mt / GWa / a"))

    # Shorthand for get_techs(rep, …)
    techs = partial(get_techs, rep)

    def full(name: str) -> Key:
        """Return the full key for `name`."""
        return single_key(rep.full_key(name))

    # L3059 from message_data/tools/post_processing/default_tables.py
    # "gas_{cc,ppl}_share": shares of gas_cc and gas_ppl in the summed output of both
    k0 = out(rep, ["gas_cc", "gas_ppl"])
    for t in "gas_cc", "gas_ppl":
        k1 = out(rep, [t])
        k2 = rep.add(Key(f"{t}_share", k1.dims), "div", k0, k1)
        assert_dims(rep, single_key(k2))

    # L3026
    # "in:*:nonccs_gas_tecs": Input to non-CCS technologies using gas at l=(secondary,
    # final), net of output from transmission and distribution technologies.
    c_gas = dict(c=["gas"])
    k0 = inp(rep, techs("gas", "all extra"), filters=c_gas)
    k1 = out(rep, ["gas_t_d", "gas_t_d_ch4"], filters=c_gas)
    k2 = rep.add(Key("in", k1.dims, "nonccs_gas_tecs"), "sub", k0, k1)
    assert_dims(rep, single_key(k2))

    # L3091
    # "Biogas_tot_abs": absolute output from t=gas_bio [energy units]
    # "Biogas_tot": above converted to its CO₂ content = CO₂ emissions from t=gas_bio
    # [mass/time]
    Biogas_tot_abs = out(rep, ["gas_bio"], name="Biogas_tot_abs")
    rep.add("Biogas_tot", "mul", Biogas_tot_abs, "crbcnt_gas", "conv_c2co2")

    # L3052
    # "in:*:all_gas_tecs": Input to all technologies using gas at l=(secondary, final),
    # including those with CCS.
    k0 = inp(
        rep,
        ["gas_cc_ccs", "meth_ng", "meth_ng_ccs", "h2_smr", "h2_smr_ccs"],
        filters=c_gas,
    )
    k1 = rep.add(
        Key("in", k0.dims, "all_gas_tecs"), "add", full("in::nonccs_gas_tecs"), k0
    )
    assert_dims(rep, k1)

    # L3165
    # "Hydrogen_tot:*": CO₂ emissions from t=h2_mix [mass/time]
    k0 = emi(
        rep,
        ["h2_mix"],
        name="_Hydrogen_tot",
        filters=dict(r=["CO2_cc"]),
        unit_key="CO2 emissions",
    )
    # NB Must alias here, otherwise full("Hydrogen_tot") below gets a larger set of
    #    dimensions than intended
    rep.add(Key("Hydrogen_tot", k0.dims), k0)

    # L3063
    # "in:*:nonccs_gas_tecs_wo_ccsretro": "in:*:nonccs_gas_tecs" minus inputs to
    # technologies fitted with CCS add-on technologies.
    filters = dict(c=["gas"], l=["secondary"])
    pe_w_ccs_retro_keys = [
        pe_w_ccs_retro(rep, *args, filters=filters)
        for args in (
            ("gas_cc", "g_ppl_co2scr", full("gas_cc_share")),
            ("gas_ppl", "g_ppl_co2scr", full("gas_ppl_share")),
            # FIXME Raises KeyError
            # ("gas_htfc", "gfc_co2scr", None),
        )
    ]
    k0 = rep.add(anon(dims=pe_w_ccs_retro_keys[0]), "add", *pe_w_ccs_retro_keys)
    k1 = rep.add(
        Key("in", k0.dims, "nonccs_gas_tecs_wo_ccsretro"),
        "sub",
        full("in::nonccs_gas_tecs"),
        k0,
    )
    assert_dims(rep, k0, k1)

    # L3144, L3234
    # "Biogas_trp", "Hydrogen_trp": transportation shares of emissions savings from
    # biogas production/use, and from hydrogen production, respectively.
    # X_trp = X_tot * (trp input of gas / `other` inputs)
    k0 = inp(rep, techs("trp gas"), filters=c_gas)
    for name, other in (
        ("Biogas", full("in::all_gas_tecs")),
        ("Hydrogen", full("in::nonccs_gas_tecs_wo_ccsretro")),
    ):
        k1 = rep.add(anon(dims=other), "div", k0, other)
        k2 = rep.add(f"{name}_trp", "mul", f"{name}_tot", k1)
        assert_dims(rep, single_key(k1))

    # L3346
    # "FE_Transport": CO₂ emissions from all transportation technologies directly using
    # fossil fuels.
    FE_Transport = emi(
        rep,
        techs("trp", "coal foil gas loil meth"),
        name="FE_Transport",
        filters=dict(r=["CO2_trp"]),
        unit_key="CO2 emissions",
    )

    # L3886
    # "Transport": CO₂ emissions from transport. "FE_Transport" minus emissions saved by
    # use of biogas in transport, plus emissions from production of hydrogen used in
    # transport.
    k0 = rep.add(anon(dims=FE_Transport), "sub", FE_Transport, full("Biogas_trp"))
    k1, *_ = iter_keys(
        rep.add(Key("Transport", k0.dims), "add", k0, full("Hydrogen_trp"), sums=True)
    )
    assert_dims(rep, k0, k1)

    # TODO Identify where to sum on "h", "m", "yv" dimensions

    # Convert to IAMC structure
    var = "Emissions|CO2|Energy|Demand|Transportation|Road Rail and Domestic Shipping"
    info = dict(variable="transport emissions", base=k1.drop("h", "m", "yv"), var=[var])
    iamc(rep, info)

    # TODO use store_ts() to store on scenario

    log.info(f"Added {len(rep.graph) - N} keys")



"""Logging utilities."""

import atexit
import logging
import logging.config
import logging.handlers
import re
import sys
import time
from contextlib import contextmanager
from datetime import datetime, timedelta, timezone
from queue import SimpleQueue
from time import process_time
from typing import Dict, Optional, Union, cast
from warnings import warn

# NB mark_time, preserve_log_level, and silence_log are exposed by util/__init__.py
__all__ = [
    "Formatter",
    "QueueListener",
    "SilenceFilter",
    "StreamHandler",
    "setup",
]

log = logging.getLogger(__name__)

# References to handlers
_HANDLER: Dict[str, logging.Handler] = dict()

# For mark_time()
_TIMES = []


class Formatter(logging.Formatter):
    """Formatter for log records.

    Parameters
    ----------
    use_color : bool, *optional*
        If :any:`True`, :mod:`colorama` is used to colour log messages.
    """

    CYAN = ""
    DIM = ""
    RESET_ALL = ""

    _short_name = None

    def __init__(self, use_colour: bool = True):
        super().__init__()

        try:
            if use_colour:
                # Import and initialize colorama
                import colorama

                colorama.init()
                self.CYAN = colorama.Fore.CYAN
                self.DIM = colorama.Style.DIM
                self.RESET_ALL = colorama.Style.RESET_ALL
        except ImportError:  # pragma: no cover
            pass  # Not installed

    def format(self, record):
        """Format `record`.

        Records are formatted like::

            model.transport.data.add_par_data  220 rows in 'input'
            ...add_par_data  further messages

        …with the calling function name (e.g. 'add_par_data') coloured for legibility
        on first occurrence, then dimmed when repeated.
        """
        # Remove the leading 'message_data.' from the module name
        name_parts = record.name.split(".")
        if name_parts[0] in ("message_ix_models", "message_data"):
            short_name = ".".join(["—"] + name_parts[1:])
        else:
            short_name = record.name

        if self._short_name != short_name:
            self._short_name = short_name
            prefix = f"{self.CYAN}{short_name}."
        else:
            prefix = f"{self.DIM}..."

        return f"{prefix}{record.funcName}{self.RESET_ALL}  {record.getMessage()}"


class QueueHandler(logging.handlers.QueueHandler):
    # For typing with Python ≤ 3.11 only; from 3.12 this attribute is described
    listener: "QueueListener"


class QueueListener(logging.handlers.QueueListener):
    """:class:`.logging.QueueListener` with a :meth:`.flush` method."""

    def flush(self):
        """Flush the queue: join the listener/monitor thread and then restart."""
        if self._thread is not None:
            super().stop()
            self.start()


class SilenceFilter(logging.Filter):
    """Log filter that only allows records from `names` that are at or above `level`."""

    __slots__ = ("level", "name_expr")

    def __init__(self, names: str, level: int):
        self.level = level
        # Compile a regular expression for the name
        self.name_re = re.compile("|".join(map(re.escape, sorted(names.split()))))

    def filter(self, record) -> bool:
        return not (record.levelno < self.level and self.name_re.match(record.name))


class StreamHandler(logging.StreamHandler):
    """Like :class:`.logging.StreamHandler`, but retrieve the stream on each access.

    This avoids the case that :mod:`click`, :mod:`pytest`, or something else adjusts
    :py:`sys.stdout` temporarily, but the handler's stored reference to the original is
    not updated.
    """

    #: Name of the :mod:`sys` stream to use, as :class:`str` rather than a direct
    #: reference.
    stream_name: str

    def __init__(self, stream_name: str):
        self.stream_name = stream_name
        logging.Handler.__init__(self)

    @property
    def stream(self):
        return getattr(sys, self.stream_name)


def mark_time(quiet: bool = False) -> None:
    """Record and log (if `quiet` is :obj:`True`) a time mark."""
    _TIMES.append(process_time())
    if not quiet and len(_TIMES) > 1:
        logging.getLogger(__name__).info(
            f" +{_TIMES[-1] - _TIMES[-2]:.1f} = {_TIMES[-1]:.1f} seconds"
        )


@contextmanager
def preserve_log_handlers(name: Optional[str] = None):
    """Context manager to preserve the handlers of a `logger`."""
    # Access the named logger
    logger = logging.getLogger(name)
    # Make a copy of its list of handlers
    handlers = list(logger.handlers)

    try:
        yield
    finally:
        # Make a list of handlers which have disappeared from the logger
        to_restore = list(filter(lambda h: h not in logger.handlers, handlers))
        for h in to_restore:
            logger.addHandler(h)
        # Log after the handlers have been restored
        log.debug(f"Restore to {logger}.handlers: {to_restore or '(none)'}")


@contextmanager
def preserve_log_level():
    """Context manager to preserve the level of the ``message_ix_models`` logger."""
    # Get the top-level logger for the package containing this file
    main_log = logging.getLogger(__name__.split(".")[0])

    try:
        # Store the current level
        level = main_log.getEffectiveLevel()
        yield
    finally:
        # Restore the level
        main_log.setLevel(level)


def configure():
    """Apply logging configuration."""
    # NB We do this programmatically as logging.config.dictConfig()'s automatic steps
    #    require adjustments that end up being more verbose and less clear.
    from platformdirs import user_log_path

    # Stream handler
    _HANDLER["console"] = h_console = StreamHandler(stream_name="stdout")
    h_console.setLevel(logging.CRITICAL)
    h_console.setFormatter(Formatter())

    # Construct the file name for the log file
    log_file_path = user_log_path("message-ix-models", ensure_exists=True).joinpath(
        datetime.now(timezone(timedelta(seconds=time.timezone)))
        .isoformat(timespec="seconds")
        .replace(":", "")
    )

    # File handler
    _HANDLER["file"] = h_file = logging.FileHandler(
        filename=str(log_file_path), delay=True
    )
    h_file.setLevel(logging.CRITICAL)
    h_file.setFormatter(Formatter(use_colour=False))

    # Queue handler
    queue = SimpleQueue()
    _HANDLER["queue"] = h_queue = QueueHandler(queue)
    logging.root.addHandler(h_queue)

    # Queue listener
    h_queue.listener = listener = QueueListener(
        queue, h_console, h_file, respect_handler_level=True
    )
    listener.start()
    atexit.register(listener.stop)

    for name, level in (
        (None, logging.DEBUG),
        # Ensure no level set for these packages; the level of the "console"/"file"
        # handlers determines outputs
        ("message_ix_models", logging.NOTSET),
        ("message_data", logging.NOTSET),
        # Hide lower-level messages for some upstream packages from the file log
        ("graphviz._tools", logging.WARNING),
        ("matplotlib", logging.WARNING),
        ("pycountry.db", logging.WARNING),
    ):
        logging.getLogger(name).setLevel(level)


def setup(
    level: Union[str, int] = 99,
    console: bool = True,
    *,
    file: bool = False,
) -> None:
    """Initialize logging.

    Parameters
    ----------
    level : str, *optional*
        Log level for the console log handler.
    console : bool, *optional*
        If :obj:`False`, do not print any messages to console.
    file : bool, *optional*
        If :obj:`False`, do not print any messages to file.
    """

    root = logging.getLogger()
    if not any(isinstance(h, logging.handlers.QueueHandler) for h in root.handlers):
        # Not yet configured
        configure()

    # Apply settings to loggers and handlers: either just-created, or pre-existing

    # Set the level of the console handler
    _HANDLER["console"].setLevel(99 if console is False else level)

    if file is False:
        _HANDLER["file"].setLevel(99)
    else:
        _HANDLER["file"].setLevel("DEBUG")
        log.info(f"Log to {cast(logging.FileHandler, _HANDLER['file']).baseFilename}")


def flush() -> None:
    """Flush the queue."""
    cast(QueueHandler, _HANDLER["queue"]).listener.flush()


@contextmanager
def silence_log(names: Optional[str] = None, level: int = logging.ERROR):
    """Context manager to temporarily quiet 1 or more loggers.

    Parameters
    ----------
    names : str, *optional*
        Space-separated names of loggers to quiet.
    level : int, *optional*
        Minimum level of log messages to allow.

    Examples
    --------
    >>> with silence_log():
    >>>     log.warning("This message is not recorded.")
    """
    if isinstance(names, list):
        warn(
            "silence_log(names=…) as list of str; use a single, space-separated str",
            DeprecationWarning,
            stacklevel=2,
        )
        names = " ".join(names)

    # Create a filter; default, the top-level logger for the current package
    f = SilenceFilter(names or f"message_data {__name__.split('.')[0]}", level)
    log.info(f"Set level={level} for logger(s): {f.name_re.pattern.replace('|', ' ')}")

    try:
        # Add the same filter to every handler of the root logger
        for handler in logging.root.handlers:
            handler.addFilter(f)

        yield
    finally:
        # Remove the filter
        for handler in logging.root.handlers:
            handler.removeFilter(f)
        log.info("…restored.")
